import json

from flask import request
from flask import Response
from sword3common.constants import JSON_LD_CONTEXT


def jsonld_serializer(data, **kwargs):
    kwargs.setdefault("mimetype", "application/ld+json")
    data = {
        "@context": JSON_LD_CONTEXT,
        "@id": data.get("@id") or request.url,
        **data,
    }
    return Response(json.dumps(data, indent=2) + "\n", **kwargs)



from http import HTTPStatus

from flask import request
from flask import Response
from invenio_db import db
from invenio_records_rest.views import need_record_permission
from invenio_records_rest.views import pass_record

from . import SWORDDepositView
from ..api import SWORDDeposit

__all__ = ["DepositMetadataView"]


class DepositMetadataView(SWORDDepositView):
    view_name = "{}_metadata"

    @pass_record
    @need_record_permission("read_permission_factory")
    def get(self, pid, record: SWORDDeposit):
        """
        Retrieve the deposit's SWORD metadata as a JSON-LD document

        This will return the empty document with a 200 status if no metadata is available
        """
        return {
            "@id": record.sword_status_url,
            **record.get("swordMetadata", {}),
        }

    @pass_record
    @need_record_permission("update_permission_factory")
    def post(self, pid, record: SWORDDeposit):
        """
        Extend the metadata with new metadata from the request body

        :param pid: The persistent identifier for the deposit
        :param record: The SWORDDeposit object
        :raises Conflict: if there is existing metadata that doesn't support the ``+`` operation
        :return: a 204 No Content response
        """
        record.set_metadata(
            request.stream, self.metadata_class, request.content_type, replace=False
        )
        record.commit()
        db.session.commit()
        return Response(status=HTTPStatus.NO_CONTENT)

    @pass_record
    @need_record_permission("update_permission_factory")
    def put(self, pid, record: SWORDDeposit):
        """
        Set the metadata with new metadata from the request body

        :param pid: The persistent identifier for the deposit
        :param record: The SWORDDeposit object
        :return: a 204 No Content response
        """
        record.set_metadata(request.stream, self.metadata_class, request.content_type)
        record.commit()
        db.session.commit()
        return Response(status=HTTPStatus.NO_CONTENT)

    @pass_record
    @need_record_permission("delete_permission_factory")
    def delete(self, pid, record: SWORDDeposit):
        """
        Delete ny existing metadata of the format given by the Metadata-Format header

        :param pid: The persistent identifier for the deposit
        :param record: The SWORDDeposit object
        :return: a 204 No Content response
        """
        record.set_metadata(None, self.metadata_class)
        record.commit()
        db.session.commit()
        return Response(status=HTTPStatus.NO_CONTENT)



from __future__ import annotations

import mimetypes
import os
import shutil
import tempfile
import uuid
import zipfile

import bagit
from invenio_files_rest.models import ObjectVersion
from sword3common.constants import PackagingFormat
from sword3common.exceptions import ContentMalformed
from sword3common.exceptions import ContentTypeNotAcceptable
from sword3common.exceptions import ValidationFailed

from ..enum import ObjectTagKey
from ..metadata import SWORDMetadata
from ..utils import TagManager
from .base import Packaging

__all__ = ["SWORDBagItPackaging"]


class SWORDBagItPackaging(Packaging):
    content_type = "application/zip"
    packaging_name = PackagingFormat.SwordBagIt

    def get_original_deposit_filename(
        self, filename: str = None, media_type: str = None
    ) -> str:
        return self.record.original_deposit_key_prefix + "bagit-{}.zip".format(
            uuid.uuid4()
        )

    def unpack(self, object_version: ObjectVersion):
        if object_version.mimetype != self.content_type:
            raise ContentTypeNotAcceptable(
                "Content-Type must be {}".format(self.content_type)
            )

        with tempfile.TemporaryDirectory() as path:
            try:
                with tempfile.TemporaryFile() as f:
                    with object_version.file.storage().open() as stream:
                        shutil.copyfileobj(stream, f)
                    f.seek(0)
                    zip = zipfile.ZipFile(f)
                    zip.extractall(path)
                    zip.close()
                    f.seek(0)

                bag = bagit.Bag(path)
                bag.validate()

                if next(bag.files_to_be_fetched(), None):
                    raise ValidationFailed("fetch.txt not supported in SWORD BagIt")

                self.record["bagitInfo"] = bag.info

                # Ingest any SWORD metadata
                metadata_path = os.path.join(path, "metadata/sword.json")
                if (
                    os.path.exists(metadata_path)
                    and "metadata/sword.json" in bag.entries
                ):
                    with open(metadata_path, "rb") as metadata_f:
                        self.record.set_metadata(
                            metadata_f,
                            metadata_class=SWORDMetadata,
                            content_type="application/ld+json",
                            derived_from=object_version.key,
                            replace=True,
                        )
                        self.record.commit()

                # Ingest payload files
                for name in bag.payload_entries():
                    with open(os.path.join(path, name), "rb") as payload_f:
                        archive_object_version = ObjectVersion.create(
                            self.record.bucket,
                            name.split(os.path.sep, 1)[-1],
                            mimetype=mimetypes.guess_type(name)[0],
                            stream=payload_f,
                        )

                        tags = TagManager(archive_object_version)
                        tags.update(
                            {
                                ObjectTagKey.FileSetFile: "true",
                                ObjectTagKey.DerivedFrom: object_version.key,
                            }
                        )
                return set(bag.payload_entries())
            except bagit.BagValidationError as e:
                raise ValidationFailed(e.message) from e
            except bagit.BagError as e:
                raise ContentMalformed(*e.args) from e
            except zipfile.BadZipFile as e:
                raise ContentMalformed("Bad ZIP file") from e



import datetime
import io
import json
import uuid
from http import HTTPStatus

import pytest

from helpers import login
from invenio_files_rest.models import MultipartObject, Part, ObjectVersion
from invenio_records.models import RecordMetadata
from invenio_sword import tasks
from invenio_sword.api import SegmentedUploadRecord, SWORDDeposit
from invenio_sword.enum import ObjectTagKey, FileState
from invenio_sword.schemas import ByReferenceFileDefinition
from invenio_sword.utils import TagManager
from sword3common.constants import JSON_LD_CONTEXT, PackagingFormat


def test_start_segmented_unauthenticated(api):
    with api.test_request_context(), api.test_client() as client:
        response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": "segment-init; segment_count=2; segment_size=1000000; size=2000000",
            },
        )
        assert response.status_code == HTTPStatus.UNAUTHORIZED


def test_start_without_content_disposition(api, users):
    with api.test_request_context(), api.test_client() as client:
        login(client)
        response = client.post("/sword/staging",)
        assert response.status_code == HTTPStatus.BAD_REQUEST


def test_start_segmented_small_segment(api, users):
    with api.test_request_context(), api.test_client() as client:
        login(client)
        response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": "segment-init; segment_count=2; segment_size=1; size=2",
            },
        )
        assert response.status_code == HTTPStatus.BAD_REQUEST
        assert "segment_size" in response.json["errors"]


def test_start_segmented(api, location, users):
    with api.test_request_context(), api.test_client() as client:
        login(client)
        response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": "segment-init; segment_count=2; segment_size=10000000; size=20000000",
            },
        )
        assert response.status_code == HTTPStatus.CREATED
        assert "Location" in response.headers

        record_metadata = RecordMetadata.query.one()
        record = SegmentedUploadRecord(record_metadata.json, model=record_metadata)

        response = client.get(response.headers["Location"])
        assert response.json == {
            "@context": JSON_LD_CONTEXT,
            "@id": "http://localhost/sword/staging/" + str(record.id),
            "@type": "Temporary",
            "segments": {
                "received": [],
                "expecting": [1, 2],
                "size": 20000000,
                "segment_size": 10000000,
            },
        }


def test_access_someone_elses_segmented_upload(api, users, location):
    with api.test_request_context(), api.test_client() as client:
        login(client, **users[0])
        response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": "segment-init; segment_count=2; segment_size=10000000; size=20000000",
            },
        )
        assert response.status_code == HTTPStatus.CREATED
        assert "Location" in response.headers

        # record_metadata = RecordMetadata.query.one()
        # record = SegmentedUploadRecord(record_metadata.json, model=record_metadata)

    with api.test_request_context(), api.test_client() as client:
        login(client, **users[1])
        response = client.get(response.headers["Location"])
        assert response.status_code == HTTPStatus.FORBIDDEN
        assert response.json["@type"] == "AuthenticationFailed"


def test_access_unknown_segmented_upload(api, users, location):
    with api.test_request_context(), api.test_client() as client:
        login(client)
        response = client.get("/sword/staging/" + str(uuid.uuid4()))
        assert response.status_code == HTTPStatus.NOT_FOUND


@pytest.mark.parametrize(
    "content_disposition",
    [
        "segment; segment_index=1",  # incorrect parameter name
        "",  # missing entirely
        "netgems; segment_number=1",  # wrong disposition type
    ],
)
def test_upload_with_wrong_content_disposition(
    api, users, location, content_disposition
):
    with api.test_request_context(), api.test_client() as client:
        login(client)
        init_response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": f"segment-init; segment_count=2; segment_size=10; size=15",
            },
        )
        assert init_response.status_code == HTTPStatus.CREATED

        response = client.post(
            init_response.headers["Location"],
            headers={"Content-Disposition": content_disposition},
            data=b"1234567890",
        )
        assert response.status_code == HTTPStatus.BAD_REQUEST


@pytest.mark.parametrize(
    "segment_count,segment_size,size",
    [
        (2, 12, 24),  # segments all the same size
        (3, 10, 26),  # differently-sized final segment
        (2, 10, 19),  # check for off-by-one errors
        (2, 10, 20),  # check for off-by-one errors
        (3, 10, 21),  # check for off-by-one errors
    ],
)
def test_upload_segments(api, users, location, segment_count, segment_size, size):
    # e.g. b'ABC...XYZ'
    data = bytes(65 + i for i in range(size))

    with api.test_request_context(), api.test_client() as client:
        login(client)
        init_response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": f"segment-init; segment_count={segment_count}; segment_size={segment_size}; size={size}",
            },
        )
        assert init_response.status_code == HTTPStatus.CREATED
        assert "Location" in init_response.headers

        segment_number = 0
        for i in range(0, size, segment_size):
            # Haven't yet finished the upload
            multipart_object: MultipartObject = MultipartObject.query.one()
            assert multipart_object.completed is False

            segment_number += 1
            response = client.post(
                init_response.headers["Location"],
                headers={
                    "Content-Disposition": f"segment; segment_number={segment_number}",
                },
                data=data[i : i + segment_size],
            )
            assert response.status_code == HTTPStatus.NO_CONTENT

        # Check that the test code got the number of segments right.
        assert segment_number == segment_count

        # MultipartObject has been marked complete
        multipart_object: MultipartObject = MultipartObject.query.one()
        assert multipart_object.completed is True


@pytest.mark.parametrize(
    "segment_number,segment_size",
    [
        (1, 8),  # too short
        (1, 12),  # too long
        (2, 3),  # too short final segment
        (2, 7),  # too long final segment
        (2, 10),  # also too long final segment, but the otherwise usual size
    ],
)
def test_upload_wrong_segment_size(api, users, location, segment_number, segment_size):
    with api.test_request_context(), api.test_client() as client:
        login(client)
        init_response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": f"segment-init; segment_count=2; segment_size=10; size=15",
            },
        )
        assert init_response.status_code == HTTPStatus.CREATED
        assert "Location" in init_response.headers

        response = client.post(
            init_response.headers["Location"],
            headers={
                "Content-Disposition": f"segment; segment_number={segment_number}",
            },
            data=bytes(65 + i for i in range(segment_size)),
        )
        assert response.status_code == HTTPStatus.BAD_REQUEST
        assert response.json["@type"] == "InvalidSegmentSize"


def test_delete_segmented_upload(api, users, location):
    with api.test_request_context(), api.test_client() as client:
        login(client)

        # Create a segmented upload
        init_response = client.post(
            "/sword/staging",
            headers={
                "Content-Disposition": f"segment-init; segment_count=2; segment_size=10; size=15",
            },
        )
        assert init_response.status_code == HTTPStatus.CREATED
        assert "Location" in init_response.headers

        # Upload a single segment
        response = client.post(
            init_response.headers["Location"],
            headers={"Content-Disposition": f"segment; segment_number=1",},
            data=b"1234567890",
        )

        # Delete the whole thing
        response = client.delete(init_response.headers["Location"],)
        assert response.status_code == HTTPStatus.NO_CONTENT

        assert MultipartObject.query.count() == 0
        assert Part.query.count() == 0


def test_post_by_reference_segmented(api, users, location, task_delay):
    with api.test_request_context(), api.test_client() as client:
        # Assemble a segmented upload from parts, and complete it
        segmented_upload_record: SegmentedUploadRecord = SegmentedUploadRecord.create(
            {}
        )
        multipart_object = MultipartObject.create(
            bucket=segmented_upload_record.bucket,
            key="some-key",
            size=15,
            chunk_size=10,
        )
        Part.create(multipart_object, 0, stream=io.BytesIO(b"abcdefghij"))
        Part.create(multipart_object, 1, stream=io.BytesIO(b"klmno"))
        multipart_object.complete()

        login(client)

        ttl = (
            datetime.datetime.now(tz=datetime.timezone.utc)
            + datetime.timedelta(0, 3600)
        ).isoformat()

        response = client.post(
            "/sword/service-document",
            data=json.dumps(
                {
                    "@context": JSON_LD_CONTEXT,
                    "@type": "ByReference",
                    "byReferenceFiles": [
                        {
                            "@id": f"http://localhost/sword/staging/{segmented_upload_record.id}",
                            "contentDisposition": "attachment; filename=some-resource.json",
                            "contentType": "application/json",
                            "dereference": True,
                            "ttl": ttl,
                        }
                    ],
                }
            ),
            headers={
                "Content-Disposition": "attachment; by-reference=true",
                "Content-Type": "application/ld+json",
            },
        )

        assert response.status_code == HTTPStatus.CREATED

        object_version = ObjectVersion.query.one()
        tags = TagManager(object_version)

        assert tags == {
            ObjectTagKey.Packaging: "http://purl.org/net/sword/3.0/package/Binary",
            ObjectTagKey.ByReferenceTemporaryID: str(segmented_upload_record.id),
            ObjectTagKey.FileState: FileState.Pending,
            ObjectTagKey.ByReferenceDereference: "true",
            ObjectTagKey.ByReferenceNotDeleted: "true",
            ObjectTagKey.OriginalDeposit: "true",
            ObjectTagKey.ByReferenceTTL: ttl,
        }


def test_by_reference_sets_tag(api, users, location, task_delay):
    with api.test_request_context():
        # Assemble a segmented upload from parts, and complete it
        segmented_upload_record: SegmentedUploadRecord = SegmentedUploadRecord.create(
            {}
        )
        multipart_object = MultipartObject.create(
            bucket=segmented_upload_record.bucket,
            key="some-key",
            size=15,
            chunk_size=10,
        )
        Part.create(multipart_object, 0, stream=io.BytesIO(b"abcdefghij"))
        Part.create(multipart_object, 1, stream=io.BytesIO(b"klmno"))
        multipart_object.complete()

        record: SWORDDeposit = SWORDDeposit.create({})
        record.set_by_reference_files(
            [
                ByReferenceFileDefinition(
                    temporary_id=segmented_upload_record.id,
                    content_disposition="attachment; filename=something.txt",
                    content_type="text/plain",
                    packaging=PackagingFormat.Binary,
                    dereference=True,
                ),
            ],
            lambda *args: True,
            "http://localhost/",
        )

        object_version = ObjectVersion.query.one()
        tags = TagManager(object_version)

        assert tags == {
            ObjectTagKey.OriginalDeposit: "true",
            ObjectTagKey.ByReferenceTemporaryID: str(segmented_upload_record.id),
            ObjectTagKey.Packaging: "http://purl.org/net/sword/3.0/package/Binary",
            ObjectTagKey.FileState: FileState.Pending,
            ObjectTagKey.ByReferenceDereference: "true",
            ObjectTagKey.ByReferenceNotDeleted: "true",
        }

        tasks.dereference_object(record.id, object_version.version_id)

        assert object_version.file.storage().open().read() == b"abcdefghijklmno"


def test_dereference_missing_upload(api, users, location, task_delay):
    with api.test_request_context():
        # Assemble a segmented upload from parts, and complete it
        segmented_upload_record: SegmentedUploadRecord = SegmentedUploadRecord.create(
            {}
        )
        MultipartObject.create(
            bucket=segmented_upload_record.bucket,
            key="some-key",
            size=15,
            chunk_size=10,
        )

        record: SWORDDeposit = SWORDDeposit.create({})
        record.set_by_reference_files(
            [
                ByReferenceFileDefinition(
                    temporary_id=segmented_upload_record.id,
                    content_disposition="attachment; filename=something.txt",
                    content_type="text/plain",
                    packaging=PackagingFormat.Binary,
                    dereference=True,
                ),
            ],
            lambda *args: True,
            "http://localhost/",
        )

        object_version = ObjectVersion.query.one()

        with pytest.raises(ValueError):
            tasks.dereference_object(record.id, object_version.version_id)

        assert TagManager(object_version)[ObjectTagKey.FileState] == FileState.Error



from . import config
from . import views


class InvenioSword(object):
    def __init__(self, app=None):
        if app:
            self.init_config(app)
            self.init_app(app)

    def init_config(self, app):
        for k in dir(config):
            if k.startswith("SWORD_"):
                app.config.setdefault(k, getattr(config, k))

    def init_app(self, app):
        app.register_blueprint(views.create_blueprint(app.config))



import os
from typing import Any
from typing import Dict

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
"""Sphinx configuration."""
# -- Project information -----------------------------------------------------
project = "Invenio-SWORD"
copyright = "2019, CottageLabs <hello@cottagelabs.com>"
author = "CottageLabs <hello@cottagelabs.com>"

# Get the version string. Cannot be done with import!
g: Dict[str, Any] = {}
with open(os.path.join("..", "invenio_sword", "version.py"), "rt") as fp:
    exec(fp.read(), g)
    release = g["__version__"]


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

html_theme_options = {
    "description": "Module for depositing record metadata and uploading files using the SWORD protocol.",
    "github_user": "swordapp",
    "github_repo": "invenio-sword",
    "github_button": False,
    "github_banner": True,
    "show_powered_by": False,
    "extra_nav_links": {
        "invenio-sword@GitHub": "https://github.com/swordapp/invenio-sword",
        # Not yet uploaded to PyPI
        # 'invenio-sword@PyPI': 'https://pypi.python.org/pypi/invenio-sword/',
        "SWORDv3 specification": "https://swordapp.github.io/swordv3/swordv3.html",
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Extension configuration -------------------------------------------------



from __future__ import annotations

import json
import typing
import uuid

import rdflib
from rdflib.namespace import DC

from ..typing import BytesReader

if typing.TYPE_CHECKING:  # pragma: nocover
    from invenio_sword.api import SWORDDeposit


class Metadata:
    content_type: str
    filename: str
    metadata_format: str

    @classmethod
    def from_document(
        cls, document: BytesReader, content_type: str, encoding: str = "utf_8"
    ) -> Metadata:
        raise NotImplementedError  # pragma: nocover

    def to_sword_metadata(self) -> dict:
        raise NotImplementedError  # pragma: nocover

    def update_record_metadata(self, record: SWORDDeposit):
        graph = rdflib.Graph()
        subject = rdflib.URIRef("urn:uuid:" + str(uuid.uuid4()))

        data = {
            **self.to_sword_metadata(),
            "@id": str(subject),
        }
        graph.parse(data=json.dumps(data), format="json-ld")

        predicates = set(graph.predicates(subject=subject))

        if "metadata" not in record:
            record["metadata"] = {}

        if DC.title in predicates:
            if "title_statement" not in record["metadata"]:
                record["metadata"]["title_statement"] = {}
            record["metadata"]["title_statement"]["title"] = str(
                graph.value(subject, DC.title)
            )
        elif "title" in record["metadata"].get("title_statement", {}):
            del record["metadata"]["title_statement"]["title"]

    def __bytes__(self):
        raise NotImplementedError  # pragma: nocover

    def __add__(self, other):
        return NotImplemented  # pragma: nocover


class JSONMetadata(Metadata):
    @classmethod
    def from_document(
        cls,
        document: typing.Union[BytesReader, dict],
        content_type: str,
        encoding: str = "utf_8",
    ) -> Metadata:
        raise NotImplementedError



from flask_login import current_user
from invenio_deposit.scopes import write_scope
from invenio_deposit.utils import check_oauth2_scope
from invenio_oauth2server import require_api_auth

from invenio_sword.api import SWORDDeposit


def is_record_owner(record: SWORDDeposit):
    return current_user.id in record["_deposit"]["owners"]


check_has_write_scope = check_oauth2_scope(lambda record: True, write_scope.id)

check_has_write_scope_and_is_record_owner = check_oauth2_scope(
    is_record_owner, write_scope.id
)


class CheckIsRecordOwner:
    def __init__(self, record):
        self.record = record

    @require_api_auth
    def can(self):
        return is_record_owner(self.record)


check_is_record_owner = CheckIsRecordOwner



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.
"""Pytest configuration."""
from __future__ import absolute_import
from __future__ import print_function

import datetime
import io
import json
import os
import shutil
import tempfile
import unittest.mock
from time import sleep

import pytest
from celery.canvas import Signature
from celery.task import Task
from elasticsearch.exceptions import RequestError
from flask import Flask
from flask_babelex import Babel
from flask_breadcrumbs import Breadcrumbs
from flask_celeryext import FlaskCeleryExt
from flask_oauthlib.provider import OAuth2Provider
from flask_security import login_user
from helpers import fill_oauth2_headers
from helpers import make_pdf_fixture
from invenio_access import InvenioAccess
from invenio_access.models import ActionUsers
from invenio_accounts import InvenioAccounts
from invenio_accounts.views import blueprint as accounts_blueprint
from invenio_assets import InvenioAssets
from invenio_db import db
from invenio_db import InvenioDB
from invenio_deposit import InvenioDeposit
from invenio_deposit import InvenioDepositREST
from invenio_deposit.api import Deposit
from invenio_deposit.scopes import write_scope
from invenio_files_rest import InvenioFilesREST
from invenio_files_rest.models import Location
from invenio_files_rest.views import blueprint as files_rest_blueprint
from invenio_files_rest.views import blueprint as invenio_files_rest_blueprint
from invenio_indexer import InvenioIndexer
from invenio_jsonschemas import InvenioJSONSchemas
from invenio_oauth2server import InvenioOAuth2Server
from invenio_oauth2server import InvenioOAuth2ServerREST
from invenio_oauth2server.models import Client
from invenio_oauth2server.models import Token
from invenio_oauth2server.views import (
    settings_blueprint as oauth2server_settings_blueprint,
)
from invenio_pidstore import InvenioPIDStore
from invenio_records import InvenioRecords
from invenio_records_rest import InvenioRecordsREST
from invenio_records_rest.utils import PIDConverter
from invenio_records_rest.views import create_blueprint_from_app as records_rest_bp
from invenio_records_ui import InvenioRecordsUI
from invenio_records_ui.views import create_blueprint_from_app as records_ui_bp
from invenio_rest import InvenioREST
from invenio_search import current_search
from invenio_search import current_search_client
from invenio_search import InvenioSearch
from invenio_search.errors import IndexAlreadyExistsError
from invenio_search_ui import InvenioSearchUI
from six import BytesIO
from six import get_method_self
from sqlalchemy import inspect
from sqlalchemy_utils.functions import create_database
from sqlalchemy_utils.functions import database_exists
from sqlalchemy_utils.functions import drop_database
from werkzeug.wsgi import DispatcherMiddleware

from invenio_sword import InvenioSword
from invenio_sword.metadata import Metadata
from invenio_sword.typing import BytesReader


def object_as_dict(obj):
    """Make a dict from SQLAlchemy object."""
    return {c.key: getattr(obj, c.key) for c in inspect(obj).mapper.column_attrs}


@pytest.fixture()
def test_metadata_format():
    return "http://example.org/TestMetadataFormat"


class TestMetadata(Metadata):
    name = "test"
    content_type = "application/binary"
    metadata_format = "http://example.org/TestMetadataFormat"
    filename = "test.bin"

    def __init__(self, data):
        self.data = data

    @classmethod
    def from_document(
        cls, document: BytesReader, content_type: str, encoding: str = "utf_8"
    ) -> Metadata:
        return cls(document.read())

    def __bytes__(self):
        return self.data


@pytest.yield_fixture()
def base_app(request, test_metadata_format):
    """Flask application fixture."""
    instance_path = tempfile.mkdtemp()

    def init_app(app_):
        app_.config.update(
            CELERY_ALWAYS_EAGER=False,
            CELERY_CACHE_BACKEND="memory",
            CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
            CELERY_RESULT_BACKEND="cache",
            JSONSCHEMAS_URL_SCHEME="http",
            SECRET_KEY="CHANGE_ME",
            SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
            SQLALCHEMY_DATABASE_URI=os.environ.get(
                "SQLALCHEMY_DATABASE_URI", "sqlite:///test.db"
            ),
            SQLALCHEMY_TRACK_MODIFICATIONS=True,
            SQLALCHEMY_ECHO=False,
            TESTING=True,
            WTF_CSRF_ENABLED=False,
            DEPOSIT_SEARCH_API="/api/search",
            SECURITY_PASSWORD_HASH="plaintext",
            SECURITY_PASSWORD_SCHEMES=["plaintext"],
            SECURITY_DEPRECATED_PASSWORD_SCHEMES=[],
            THEME_SITENAME="Test Site",
            OAUTHLIB_INSECURE_TRANSPORT=True,
            OAUTH2_CACHE_TYPE="simple",
            ACCOUNTS_JWT_ENABLE=False,
            # This allows access to files across all of invenio-files-rest
            FILES_REST_PERMISSION_FACTORY=lambda *a, **kw: type(
                "Allow", (object,), {"can": lambda self: True}
            )(),
            FILES_REST_MULTIPART_CHUNKSIZE_MIN=10,
        )
        Babel(app_)
        FlaskCeleryExt(app_)
        Breadcrumbs(app_)
        OAuth2Provider(app_)
        InvenioDB(app_)
        InvenioAccounts(app_)
        InvenioAccess(app_)
        InvenioIndexer(app_)
        InvenioJSONSchemas(app_)
        InvenioOAuth2Server(app_)
        InvenioPIDStore(app_)
        InvenioRecords(app_)
        search = InvenioSearch(app_)
        search.register_mappings("deposits", "invenio_deposit.mappings")

    api_app = Flask("testapiapp", instance_path=instance_path)
    api_app.url_map.converters["pid"] = PIDConverter
    # initialize InvenioDeposit first in order to detect any invalid dependency
    InvenioDepositREST(api_app)

    init_app(api_app)
    InvenioREST(api_app)
    InvenioOAuth2ServerREST(api_app)
    InvenioRecordsREST(api_app)
    InvenioFilesREST(api_app)
    InvenioSword(api_app)
    api_app.register_blueprint(files_rest_blueprint)
    # api_app.register_blueprint(records_files_bp(api_app))
    api_app.register_blueprint(records_rest_bp(api_app))
    api_app.register_blueprint(invenio_files_rest_blueprint)

    # Register a test (alternate) metadata format
    api_app.config["SWORD_ENDPOINTS"]["depid"]["metadata_formats"][
        test_metadata_format
    ] = TestMetadata

    app = Flask("testapp", instance_path=instance_path)
    app.url_map.converters["pid"] = PIDConverter
    # initialize InvenioDeposit first in order to detect any invalid dependency
    InvenioDeposit(app)
    init_app(app)
    app.register_blueprint(accounts_blueprint)
    app.register_blueprint(oauth2server_settings_blueprint)
    InvenioAssets(app)
    InvenioSearchUI(app)
    InvenioRecordsUI(app)
    app.register_blueprint(records_ui_bp(app))
    app.wsgi_app = DispatcherMiddleware(app.wsgi_app, {"/api": api_app.wsgi_app})

    with app.app_context():
        if str(db.engine.url) != "sqlite://" and not database_exists(
            str(db.engine.url)
        ):
            create_database(str(db.engine.url))
        db.create_all()

    yield app

    with app.app_context():
        if str(db.engine.url) != "sqlite://":
            drop_database(str(db.engine.url))
        shutil.rmtree(instance_path)


@pytest.yield_fixture()
def app(base_app):
    """Yield the REST API application in its context."""
    with base_app.app_context():
        yield base_app


@pytest.yield_fixture()
def api(base_app):
    """Yield the REST API application in its context."""
    api = get_method_self(base_app.wsgi_app.mounts["/api"])
    # api.register_blueprint(records_rest_bp(api))
    with api.app_context():
        yield api


@pytest.yield_fixture()
def test_client(base_app):
    """Test client."""
    with base_app.test_client() as client_:
        yield client_


@pytest.fixture()
def users(app):
    """Create users."""
    with db.session.begin_nested():
        datastore = app.extensions["security"].datastore
        user1 = datastore.create_user(
            email="info@inveniosoftware.org", password="tester", active=True
        )
        user2 = datastore.create_user(
            email="test@inveniosoftware.org", password="tester2", active=True
        )
        admin = datastore.create_user(
            email="admin@inveniosoftware.org", password="tester3", active=True
        )
        # Assign deposit-admin-access to admin only.
        db.session.add(ActionUsers(action="deposit-admin-access", user=admin))
    db.session.commit()
    return [object_as_dict(user1), object_as_dict(user2)]


@pytest.fixture()
def client(app, users):
    """Create client."""
    with db.session.begin_nested():
        # create resource_owner -> client_1
        client_ = Client(
            client_id="client_test_u1c1",
            client_secret="client_test_u1c1",
            name="client_test_u1c1",
            description="",
            is_confidential=False,
            user_id=users[0]["id"],
            _redirect_uris="",
            _default_scopes="",
        )
        db.session.add(client_)
    db.session.commit()
    return client_


@pytest.fixture()
def write_token_user_1(app, client, users):
    """Create token."""
    with db.session.begin_nested():
        token_ = Token(
            client=client,
            user_id=users[0]["id"],
            access_token="dev_access_1",
            refresh_token="dev_refresh_1",
            expires=datetime.datetime.now() + datetime.timedelta(hours=10),
            is_personal=False,
            is_internal=True,
            _scopes=write_scope.id,
        )
        db.session.add(token_)
    db.session.commit()
    return token_


@pytest.fixture()
def write_token_user_2(app, client, users):
    """Create token."""
    with db.session.begin_nested():
        token_ = Token(
            client=client,
            user_id=users[1]["id"],
            access_token="dev_access_2",
            refresh_token="dev_refresh_2",
            expires=datetime.datetime.now() + datetime.timedelta(hours=10),
            is_personal=False,
            is_internal=True,
            _scopes=write_scope.id,
        )
        db.session.add(token_)
    db.session.commit()
    return token_


@pytest.fixture()
def fake_schemas(app, api, es, tmpdir):
    """Fake schema."""
    schemas = tmpdir.mkdir("schemas")
    empty_schema = '{"title": "Empty"}'
    for path in (
        ("deposit-v1.0.0.json",),
        ("deposits", "test-v1.0.0.json"),
        ("test-v1.0.0.json",),
    ):
        schema = schemas
        for section in path[:-1]:
            schema = schema.mkdir(section)
        schema = schema.join(path[-1])
        schema.write(empty_schema)

    app.extensions["invenio-jsonschemas"].register_schemas_dir(schemas.strpath)
    api.extensions["invenio-jsonschemas"].register_schemas_dir(schemas.strpath)


@pytest.yield_fixture()
def es(app):
    """Elasticsearch fixture."""
    try:
        list(current_search.create())
    except (RequestError, IndexAlreadyExistsError):
        list(current_search.delete(ignore=[404]))
        list(current_search.create(ignore=[400]))
    current_search_client.indices.refresh()
    yield current_search_client
    list(current_search.delete(ignore=[404]))


@pytest.fixture()
def location(app):
    """Create default location."""
    tmppath = tempfile.mkdtemp()
    with db.session.begin_nested():
        Location.query.delete()
        loc = Location(name="local", uri=tmppath, default=True)
        db.session.add(loc)
    db.session.commit()
    return location


@pytest.fixture()
def deposit(app, es, users, location):
    """New deposit with files."""
    record = {"title": "fuu"}
    with app.test_request_context():
        datastore = app.extensions["security"].datastore
        login_user(datastore.find_user(email=users[0]["email"]))
        deposit = Deposit.create(record)
        deposit.commit()
        db.session.commit()
    sleep(2)
    return deposit


@pytest.fixture()
def files(app, deposit):
    """Add a file to the deposit."""
    content = b"### Testing textfile ###"
    stream = BytesIO(content)
    key = "hello.txt"
    deposit.files[key] = stream
    deposit.commit()
    db.session.commit()
    return list(deposit.files)


@pytest.fixture()
def pdf_file(app):
    """Create a test pdf file."""
    return {"file": make_pdf_fixture("test.pdf"), "name": "test.pdf"}


@pytest.fixture()
def pdf_file2(app):
    """Create a test pdf file."""
    return {"file": make_pdf_fixture("test2.pdf", "test"), "name": "test2.pdf"}


@pytest.fixture()
def pdf_file2_samename(app):
    """Create a test pdf file."""
    return {"file": make_pdf_fixture("test2.pdf", "test same"), "name": "test2.pdf"}


@pytest.fixture()
def json_headers():
    """JSON headers."""
    return [("Content-Type", "application/json"), ("Accept", "application/json")]


@pytest.fixture()
def oauth2_headers_user_1(app, json_headers, write_token_user_1):
    """Authentication headers (with a valid oauth2 token).

    It uses the token associated with the first user.
    """
    return fill_oauth2_headers(json_headers, write_token_user_1)


@pytest.fixture()
def oauth2_headers_user_2(app, json_headers, write_token_user_2):
    """Authentication headers (with a valid oauth2 token).

    It uses the token associated with the second user.
    """
    return fill_oauth2_headers(json_headers, write_token_user_2)


@pytest.fixture()
def metadata_document():
    return io.BytesIO(
        json.dumps(
            {
                "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                "@id": "http://example.com/object/1/metadata",
                "@type": "Metadata",
                "dc:title": "The title",
                "dcterms:abstract": "This is my abstract",
                "dc:contributor": "A.N. Other",
            }
        ).encode()
    )


@pytest.yield_fixture()
def task_delay():
    mock_obj = unittest.mock.Mock()

    def delay(self):
        mock_obj(self)

    with unittest.mock.patch.object(
        Signature, "delay", delay
    ), unittest.mock.patch.object(Task, "delay", delay):
        yield mock_obj


@pytest.fixture()
def fixtures_path():
    return os.path.join(os.path.dirname(__file__), "fixtures")



from __future__ import annotations

import mimetypes
import shutil
import tempfile
import uuid
import zipfile

from invenio_files_rest.models import ObjectVersion
from sword3common.constants import PackagingFormat
from sword3common.exceptions import ContentMalformed
from sword3common.exceptions import ContentTypeNotAcceptable

from ..enum import ObjectTagKey
from ..utils import TagManager
from .base import Packaging


__all__ = ["SimpleZipPackaging"]


class SimpleZipPackaging(Packaging):
    content_type = "application/zip"
    packaging_name = PackagingFormat.SimpleZip

    def get_original_deposit_filename(
        self, filename: str = None, media_type: str = None
    ) -> str:
        return self.record.original_deposit_key_prefix + "simple-zip-{}.zip".format(
            uuid.uuid4()
        )

    def unpack(self, object_version: ObjectVersion):
        if object_version.mimetype != self.content_type:
            raise ContentTypeNotAcceptable(
                "Content-Type must be {}".format(self.content_type)
            )

        try:
            with tempfile.TemporaryFile() as f:
                with object_version.file.storage().open() as stream:
                    shutil.copyfileobj(stream, f)
                f.seek(0)

                with zipfile.ZipFile(f) as zip:
                    names = set(zip.namelist())

                    for name in names:
                        archive_object_version = ObjectVersion.create(
                            self.record.bucket,
                            name,
                            mimetype=mimetypes.guess_type(name)[0],
                            stream=zip.open(name),
                        )

                        tags = TagManager(archive_object_version)
                        tags.update(
                            {
                                ObjectTagKey.FileSetFile: "true",
                                ObjectTagKey.DerivedFrom: object_version.key,
                            }
                        )
                return names
        except zipfile.BadZipFile as e:
            raise ContentMalformed("Bad ZIP file") from e



import io
import json
import unittest.mock
import urllib.error
import uuid
from http import HTTPStatus

import pytest
import pytest_httpserver
from flask_security import url_for_security
from invenio_db import db
from invenio_files_rest.models import Bucket
from invenio_files_rest.models import ObjectVersion
from invenio_records.models import RecordMetadata
from invenio_sword.schemas import ByReferenceSchema
from sword3common.constants import JSON_LD_CONTEXT
from sword3common.constants import PackagingFormat
from sword3common.exceptions import ContentTypeNotAcceptable

from invenio_sword import tasks
from invenio_sword.api import SWORDDeposit
from invenio_sword.enum import FileState
from invenio_sword.enum import ObjectTagKey
from invenio_sword.utils import TagManager


@pytest.mark.parametrize(
    "data,fields_with_errors, fields_without_errors",
    [
        # Missing fields
        ({}, ["@context", "@type", "byReferenceFiles"], []),
        # Wrong @context value
        ({"@context": "http://example.com/"}, ["@context"], []),
        # Correct @context value
        ({"@context": JSON_LD_CONTEXT}, [], ["@context"]),
        # Wrong @type value
        ({"@type": "Metadata"}, ["@type"], []),
        # Correct @type value
        ({"@type": "ByReference"}, [], ["@type"]),
        # Missing fields in byReferenceFiles
        ({"byReferenceFiles": [{}]}, ["byReferenceFiles"], []),
    ],
)
def test_by_reference_validation(
    api, users, location, es, data, fields_with_errors, fields_without_errors
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post(
            "/sword/service-document",
            data=json.dumps(data),
            headers={
                "Content-Disposition": "attachment; by-reference=true",
                "Content-Type": "application/ld+json",
            },
        )
        assert response.status_code == HTTPStatus.BAD_REQUEST
        assert response.is_json
        assert response.json["@type"] == "ValidationFailed"
        # The fields with errors are the superset of the ones we expected; i.e. it shouldn't accept data we
        # in this test know is wrong
        print(response.json)
        assert set(response.json["errors"]) >= set(fields_with_errors)
        # We know these fields to be good
        assert not (set(response.json["errors"]) & set(fields_without_errors))


def test_by_reference_deposit(
    api,
    users,
    location,
    es,
    httpserver: pytest_httpserver.HTTPServer,
    task_delay: unittest.mock.Mock,
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post(
            "/sword/service-document",
            data=json.dumps(
                {
                    "@context": JSON_LD_CONTEXT,
                    "@type": "ByReference",
                    "byReferenceFiles": [
                        {
                            "@id": httpserver.url_for("some-resource.json"),
                            "contentDisposition": "attachment; filename=some-resource.json",
                            "contentType": "application/json",
                            "dereference": True,
                        }
                    ],
                }
            ),
            headers={
                "Content-Disposition": "attachment; by-reference=true",
                "Content-Type": "application/ld+json",
            },
        )
        assert response.status_code == HTTPStatus.CREATED
        assert response.is_json

        bucket = Bucket.query.one()
        object_version: ObjectVersion = ObjectVersion.query.filter(
            ObjectVersion.bucket == bucket
        ).one()

        record_metadata = RecordMetadata.query.one()

        assert task_delay.call_args_list == (
            [
                unittest.mock.call(
                    tasks.dereference_object.s(
                        str(record_metadata.id), str(object_version.version_id)
                    )
                    | tasks.unpack_object.si(
                        str(record_metadata.id), str(object_version.version_id)
                    )
                )
            ]
        )

        # Ensure that no requests were made
        assert httpserver.log == []


def test_by_reference_and_metadata_deposit(
    api,
    users,
    location,
    es,
    httpserver: pytest_httpserver.HTTPServer,
    task_delay: unittest.mock.Mock,
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post(
            "/sword/service-document",
            data=json.dumps(
                {
                    "metadata": {
                        "@context": JSON_LD_CONTEXT,
                        "@type": "Metadata",
                        "dc:title": "Some data",
                    },
                    "by-reference": {
                        "@context": JSON_LD_CONTEXT,
                        "@type": "ByReference",
                        "byReferenceFiles": [
                            {
                                "@id": httpserver.url_for("some-resource.json"),
                                "contentDisposition": "attachment; filename=some-resource.json",
                                "contentType": "application/json",
                                "dereference": True,
                            }
                        ],
                    },
                }
            ),
            headers={
                "Content-Disposition": "attachment; by-reference=true; metadata=true",
                "Content-Type": "application/ld+json",
            },
        )
        assert response.status_code == HTTPStatus.CREATED
        assert response.is_json

        bucket = Bucket.query.one()
        object_version: ObjectVersion = ObjectVersion.query.filter(
            ObjectVersion.bucket == bucket, ObjectVersion.key == "some-resource.json",
        ).one()

        record_metadata = RecordMetadata.query.one()

        assert record_metadata.json["metadata"] == {
            "title_statement": {"title": "Some data"}
        }

        assert task_delay.call_args_list == (
            [
                unittest.mock.call(
                    tasks.dereference_object.s(
                        str(record_metadata.id), str(object_version.version_id)
                    )
                    | tasks.unpack_object.si(
                        str(record_metadata.id), str(object_version.version_id)
                    )
                )
            ]
        )

        # Ensure that no requests were made
        assert httpserver.log == []


def test_dereference_task(
    api, users, location, es, httpserver: pytest_httpserver.HTTPServer
):
    file_contents = "File contents.\n"

    with api.test_request_context():
        record = SWORDDeposit.create({})
        object_version = ObjectVersion.create(bucket=record.bucket, key="some-file.txt")
        TagManager(object_version).update(
            {
                ObjectTagKey.ByReferenceURL: httpserver.url_for("some-file.txt"),
                # This one should get removed after dereferencing
                ObjectTagKey.ByReferenceNotDeleted: "true",
                ObjectTagKey.Packaging: PackagingFormat.SimpleZip,
            }
        )

        httpserver.expect_request("/some-file.txt").respond_with_data(file_contents)

        db.session.refresh(object_version)
        tasks.dereference_object(record.id, object_version.version_id)

        # Check requests
        assert len(httpserver.log) == 1
        assert httpserver.log[0][0].path == "/some-file.txt"

        db.session.refresh(object_version)
        assert object_version.file is not None
        assert object_version.file.storage().open().read() == file_contents.encode(
            "utf-8"
        )

        assert TagManager(object_version) == {
            ObjectTagKey.ByReferenceURL: httpserver.url_for("some-file.txt"),
            ObjectTagKey.Packaging: PackagingFormat.SimpleZip,
            ObjectTagKey.FileState: FileState.Pending,
        }


def test_dereference_without_url(api, location, es):
    with api.test_request_context():
        record = SWORDDeposit.create({})

        object_version = ObjectVersion.create(bucket=record.bucket, key="some-file.txt")
        with pytest.raises(ValueError):
            tasks.dereference_object(record.id, object_version.version_id)


def test_dereference_already_dereferenced(
    api, location, es, httpserver: pytest_httpserver.HTTPServer
):
    with api.test_request_context():
        record = SWORDDeposit.create({})

        object_version = ObjectVersion.create(
            bucket=record.bucket, key="some-file.txt", stream=io.BytesIO(b"data")
        )
        TagManager(object_version).update(
            {
                ObjectTagKey.ByReferenceURL: httpserver.url_for("some-file.txt"),
                ObjectTagKey.Packaging: PackagingFormat.SimpleZip,
            }
        )

        httpserver.expect_request("/some-file.txt").respond_with_data(b"data")

        db.session.refresh(object_version)

        result = tasks.dereference_object(record.id, object_version.version_id)
        assert result == ["some-file.txt"]

        assert httpserver.log == []


def test_dereference_when_not_head(
    api, location, es, httpserver: pytest_httpserver.HTTPServer
):
    with api.test_request_context():
        record = SWORDDeposit.create({})

        object_version = ObjectVersion.create(
            bucket=record.bucket, key="some-file.txt", stream=io.BytesIO(b"data")
        )
        TagManager(object_version).update(
            {
                ObjectTagKey.ByReferenceURL: httpserver.url_for("some-file.txt"),
                ObjectTagKey.Packaging: PackagingFormat.SimpleZip,
            }
        )
        # This makes the object version we already had a non-head one
        ObjectVersion.delete(record.bucket, object_version.key)

        httpserver.expect_request("/some-file.txt").respond_with_data(b"data")

        db.session.refresh(object_version)

        result = tasks.dereference_object(record.id, object_version.version_id)
        assert result == ["some-file.txt"]
        assert httpserver.log == []


def test_error_dereferencing(
    api, users, location, es, httpserver: pytest_httpserver.HTTPServer
):
    with api.test_request_context():
        record = SWORDDeposit.create({})
        object_version = ObjectVersion.create(bucket=record.bucket, key="some-file.txt")
        TagManager(object_version).update(
            {
                ObjectTagKey.ByReferenceURL: httpserver.url_for("some-file.txt"),
                # This one should get removed after dereferencing
                ObjectTagKey.ByReferenceNotDeleted: "true",
                ObjectTagKey.Packaging: PackagingFormat.SimpleZip,
            }
        )

        httpserver.expect_request("/some-file.txt").respond_with_data(
            b"", status=HTTPStatus.GONE
        )

        db.session.refresh(object_version)

        with pytest.raises(urllib.error.HTTPError):
            tasks.dereference_object(record.id, object_version.version_id)

        db.session.refresh(object_version)

        tags = TagManager(object_version)
        assert tags.get(ObjectTagKey.FileState) == FileState.Error


def test_error_unpacking(api, users, location, es):
    with api.test_request_context():
        record = SWORDDeposit.create({})
        object_version = ObjectVersion.create(
            bucket=record.bucket, key="some-file.txt", mimetype="text/plain"
        )
        TagManager(object_version).update(
            {ObjectTagKey.Packaging: PackagingFormat.SimpleZip,}
        )

        db.session.refresh(object_version)

        with pytest.raises(ContentTypeNotAcceptable):
            tasks.unpack_object(record.id, object_version.version_id)

        db.session.refresh(object_version)

        tags = TagManager(object_version)
        assert tags.get(ObjectTagKey.FileState) == FileState.Error


def test_schema_different_url_types(api):
    temporary_id = uuid.uuid4()

    schema = ByReferenceSchema(
        context={"url_adapter": api.url_map.bind("invenio.example")}
    )
    result = schema.load(
        {
            "@context": JSON_LD_CONTEXT,
            "@type": "ByReference",
            "byReferenceFiles": [
                {
                    "@id": "http://elsewhere.example/",
                    "contentDisposition": "attachment; filename=file.txt",
                    "contentType": "text/plain",
                    "dereference": True,
                },
                {
                    "@id": "http://invenio.example/sword/service-document",
                    "contentDisposition": "attachment; filename=file.txt",
                    "contentType": "text/plain",
                    "dereference": True,
                },
                {
                    "@id": f"http://invenio.example/sword/staging/{temporary_id}",
                    "contentDisposition": "attachment; filename=file.txt",
                    "contentType": "text/plain",
                    "dereference": True,
                },
            ],
        }
    )

    assert result["files"][0].url == "http://elsewhere.example/"
    assert result["files"][1].url == "http://invenio.example/sword/service-document"
    assert result["files"][2].temporary_id == temporary_id



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type
from typing import Union

import sys


if TYPE_CHECKING:  # pragma: nocover
    from invenio_files_rest.models import ObjectVersion
    from invenio_sword.metadata import Metadata
    from invenio_sword.packaging import Packaging
    from invenio_sword.schemas import ByReferenceFileDefinition
    from invenio_deposit.search import DepositSearch
    from invenio_sword.api import SWORDDeposit

if sys.version_info >= (3, 8):  # pragma: nocover
    from typing import Protocol
    from typing import runtime_checkable
    from typing import TypedDict
else:  # pragma: nocover
    from typing_extensions import Protocol  # type: ignore
    from typing_extensions import runtime_checkable
    from typing_extensions import TypedDict


@runtime_checkable
class BytesReader(Protocol):
    def read(self, amount: int = -1) -> bytes:
        ...  # pragma: nocover


class SwordEndpointDefinition(TypedDict):
    packaging_formats: Dict[str, Type[Packaging]]
    metadata_formats: Dict[str, Type[Metadata]]
    default_packaging_format: str
    default_metadata_format: str

    record_class: Optional[Union[Type[SWORDDeposit], str]]
    search_class: Optional[Union[Type[DepositSearch], str]]
    search_index: Optional[str]
    search_type: Optional[Any]
    indexer_class: Any

    create_permission_factory_imp: Callable
    read_permission_factory_imp: Callable
    update_permission_factory_imp: Callable
    delete_permission_factory_imp: Callable

    service_document_route: str
    item_route: str
    metadata_route: str
    fileset_route: str
    file_route: str

    default_media_type: str

    dereference_policy: Callable[[ObjectVersion, ByReferenceFileDefinition], bool]



import io

from invenio_sword.schemas import ByReferenceFileDefinition
from sword3common.constants import PackagingFormat

from invenio_sword import tasks
from invenio_sword.api import SWORDDeposit


def test_delete_old_files(api, location, es, task_delay):
    with api.test_request_context():
        record: SWORDDeposit = SWORDDeposit.create({})

        record.set_by_reference_files(
            [
                ByReferenceFileDefinition(
                    url="http://example.com/one",
                    content_disposition="attachment; filename=br-yes.html",
                    content_type="text/html",
                    content_length=100,
                    packaging=PackagingFormat.Binary,
                    dereference=False,
                ),
                ByReferenceFileDefinition(
                    url="http://example.com/two",
                    content_disposition="attachment; filename=br-no.html",
                    content_type="text/html",
                    packaging=PackagingFormat.Binary,
                    dereference=False,
                ),
            ],
            dereference_policy=lambda record, brf: brf.dereference,
            request_url="http://localhost/something",
            replace=False,
        )

        record.ingest_file(
            io.BytesIO(b"data"),
            packaging_name=PackagingFormat.Binary,
            content_type="text/html",
            content_disposition="attachment; filename=direct-yes.html",
            replace=False,
        )
        record.ingest_file(
            io.BytesIO(b"data"),
            packaging_name=PackagingFormat.Binary,
            content_type="text/html",
            content_disposition="attachment; filename=direct-no.html",
            replace=False,
        )

        assert sorted(file.key for file in record.files) == [
            "br-no.html",
            "br-yes.html",
            "direct-no.html",
            "direct-yes.html",
        ]

        tasks.delete_old_objects(
            ["br-yes.html", "direct-yes.html"], bucket_id=record.bucket_id
        )

        assert sorted(file.key for file in record.files) == [
            "br-yes.html",
            "direct-yes.html",
        ]



from __future__ import annotations

import json
import typing

from sword3common.constants import MetadataFormat
from sword3common.exceptions import ContentMalformed
from sword3common.exceptions import ContentTypeNotAcceptable

from .base import JSONMetadata
from invenio_sword.typing import BytesReader

__all__ = ["SWORDMetadata"]


class SWORDMetadata(JSONMetadata):
    content_type = "application/ld+json"
    filename = "sword.jsonld"
    metadata_format = MetadataFormat.Sword

    def __init__(self, data):
        self.data = data

    @classmethod
    def from_document(
        cls,
        document: typing.Union[BytesReader, dict],
        content_type: str,
        encoding: str = "utf_8",
    ) -> SWORDMetadata:
        if content_type not in (cls.content_type, "application/json"):
            raise ContentTypeNotAcceptable(
                "Content-Type must be {}".format(cls.content_type)
            )
        if isinstance(document, BytesReader):
            try:
                data = json.load(document, encoding=encoding)
            except json.JSONDecodeError as e:
                raise ContentMalformed("Unable to parse JSON") from e
        else:
            data = document
        data.pop("@id", None)
        return cls(data)

    def to_sword_metadata(self) -> dict:
        # Record the full SWORD metadata without the '@id' key
        return {k: self.data[k] for k in self.data if k != "@id"}

    def __add__(self, other):
        if not isinstance(other, SWORDMetadata):
            return NotImplemented
        return type(self)({**self.data, **other.data})

    def __bytes__(self):
        return json.dumps(self.data, indent=2).encode("utf-8")



import datetime
import json
import typing
from copy import deepcopy
from functools import partial
from http import HTTPStatus

import marshmallow.exceptions
from flask import Blueprint
from flask import Response

import sword3common
from invenio_deposit.search import DepositSearch
from invenio_deposit.views.rest import create_error_handlers
from invenio_records_rest.utils import obj_or_import_string

from . import (
    DepositFilesetView,
    DepositFileView,
    DepositMetadataView,
    DepositStatusView,
    ServiceDocumentView,
    StagingURLView,
    TemporaryURLView,
)

from .. import serializers
from ..api import SWORDDeposit
from ..typing import SwordEndpointDefinition

__all__ = ["create_blueprint"]


def create_blueprint(config) -> Blueprint:
    """
    Create an Invenio-SWORD blueprint

    This takes a list of endpoint definitions and returns a SWORD blueprint for use in Invenio. You shouldn't need to
    use this directly, as it's called by :class:`invenio_sword.ext.InvenioSword`.

    See: :data:`invenio_sword.config.SWORD_ENDPOINTS`.

    :param endpoints: List of endpoints configuration.
    :returns: The configured blueprint.
    """

    endpoints: typing.Dict[str, SwordEndpointDefinition] = config["SWORD_ENDPOINTS"]

    blueprint = Blueprint("invenio_sword", __name__, url_prefix="",)
    create_error_handlers(blueprint)

    for endpoint, options in (endpoints or {}).items():
        options = deepcopy(options)

        options.setdefault("search_class", DepositSearch)
        search_class = obj_or_import_string(options["search_class"])

        # records rest endpoints will use the deposit class as record class
        options.setdefault("record_class", SWORDDeposit)
        record_class = obj_or_import_string(options["record_class"])

        # backward compatibility for indexer class
        options.setdefault("indexer_class", None)

        search_class_kwargs = {}
        if options.get("search_index"):
            search_class_kwargs["index"] = options["search_index"]

        if options.get("search_type"):
            search_class_kwargs["doc_type"] = options["search_type"]

        ctx = dict(
            read_permission_factory=obj_or_import_string(
                options.get("read_permission_factory_imp")
            ),
            create_permission_factory=obj_or_import_string(
                options.get("create_permission_factory_imp")
            ),
            update_permission_factory=obj_or_import_string(
                options.get("update_permission_factory_imp")
            ),
            delete_permission_factory=obj_or_import_string(
                options.get("delete_permission_factory_imp")
            ),
            record_class=record_class,
            search_class=partial(search_class, **search_class_kwargs),
            default_media_type=options.get("default_media_type"),
            pid_type=endpoint,
        )

        blueprint.add_url_rule(
            options["service_document_route"],
            endpoint=ServiceDocumentView.view_name.format(endpoint),
            view_func=ServiceDocumentView.as_view(
                "service",
                serializers={"application/ld+json": serializers.jsonld_serializer,},
                ctx=ctx,
            ),
        )

        blueprint.add_url_rule(
            options["item_route"],
            endpoint=DepositStatusView.view_name.format(endpoint),
            view_func=DepositStatusView.as_view(
                "item",
                serializers={"application/ld+json": serializers.jsonld_serializer,},
                ctx=ctx,
            ),
        )
        blueprint.add_url_rule(
            options["metadata_route"],
            endpoint=DepositMetadataView.view_name.format(endpoint),
            view_func=DepositMetadataView.as_view(
                "metadata",
                serializers={"application/ld+json": serializers.jsonld_serializer,},
                ctx=ctx,
            ),
        )
        blueprint.add_url_rule(
            options["fileset_route"],
            endpoint=DepositFilesetView.view_name.format(endpoint),
            view_func=DepositFilesetView.as_view(
                "fileset",
                serializers={"application/ld+json": serializers.jsonld_serializer,},
                ctx=ctx,
            ),
        )
        blueprint.add_url_rule(
            options["file_route"],
            endpoint=DepositFileView.view_name.format(endpoint),
            view_func=DepositFileView.as_view(
                name="file",
                serializers={
                    "*/*": lambda *args, **kwargs: Response(
                        status=HTTPStatus.NO_CONTENT
                    )
                },
            ),
        )

    blueprint.add_url_rule(
        config["SWORD_STAGING_URL_ROUTE"],
        endpoint=StagingURLView.view_name,
        view_func=StagingURLView.as_view(
            "staging-url",
            serializers={"application/ld+json": serializers.jsonld_serializer,},
            ctx=config["SWORD_SEGMENTED_UPLOAD_CONTEXT"],
        ),
    )

    blueprint.add_url_rule(
        config["SWORD_TEMPORARY_URL_ROUTE"],
        endpoint=TemporaryURLView.view_name,
        view_func=TemporaryURLView.as_view(
            "temporary-url",
            serializers={"application/ld+json": serializers.jsonld_serializer,},
            ctx=config["SWORD_SEGMENTED_UPLOAD_CONTEXT"],
        ),
    )

    @blueprint.errorhandler(sword3common.exceptions.SwordException)
    def sword_exception_handler(exc):
        return Response(
            json.dumps(
                {
                    "@context": sword3common.constants.JSON_LD_CONTEXT,
                    "@type": exc.name,
                    "error": exc.reason,
                    "log": exc.message,
                    "timestamp": exc.timestamp.isoformat(),
                }
            ),
            content_type="application/ld+json",
            status=exc.status_code,
        )

    @blueprint.errorhandler(marshmallow.exceptions.ValidationError)
    def marshmallow_exception_handler(exc):
        return Response(
            json.dumps(
                {
                    "@context": sword3common.constants.JSON_LD_CONTEXT,
                    "@type": sword3common.exceptions.ValidationFailed.name,
                    "error": sword3common.exceptions.ValidationFailed.reason,
                    "timestamp": datetime.datetime.now(
                        tz=datetime.timezone.utc
                    ).isoformat(),
                    "errors": exc.messages,
                }
            ),
            content_type="application/ld+json",
            status=sword3common.exceptions.ValidationFailed.status_code,
        )

    return blueprint



import os
from http import HTTPStatus

from flask_security import url_for_security
from invenio_db import db
from invenio_files_rest.models import ObjectVersion

from invenio_sword.api import SWORDDeposit
from invenio_sword.packaging import SWORDBagItPackaging


def create_bagit_record(fixtures_path):
    with open(os.path.join(fixtures_path, "bagit.zip"), "rb") as f:
        record = SWORDDeposit.create({})
        packaging = SWORDBagItPackaging(record)
        object_version = ObjectVersion.create(
            bucket=record.bucket,
            key=packaging.get_original_deposit_filename(),
            stream=f,
        )
        packaging.unpack(object_version)
        record.commit()
        db.session.commit()
    return record


def test_file_get(fixtures_path, location, es, api, users):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        bagit_record = create_bagit_record(fixtures_path)

        status_document = bagit_record.get_status_as_jsonld()
        assert len(status_document["links"]) > 0

        expected_content = []
        for file in bagit_record.files:
            with file.obj.file.storage().open() as f:
                expected_content.append(f.read())
        expected_content.sort()

        actual_content = []
        for link in status_document["links"]:
            response = client.get(link["@id"])
            assert response.status_code == HTTPStatus.OK
            actual_content.append(response.data)
        actual_content.sort()

        assert expected_content == actual_content



from __future__ import annotations

import mimetypes
import typing
import uuid
from typing import Collection
from typing import Union

from flask import current_app
from invenio_files_rest.models import ObjectVersion

if typing.TYPE_CHECKING:  # pragma: nocover
    from ..api import SWORDDeposit


class Packaging:
    packaging_name: str

    def __init__(self, record: SWORDDeposit):
        self.record = record

    def get_original_deposit_filename(
        self, filename: str = None, media_type: str = None
    ) -> str:
        if filename:
            return filename
        elif media_type and mimetypes.guess_extension(media_type):
            return "{}{}".format(uuid.uuid4(), mimetypes.guess_extension(media_type))
        else:
            return "{}.bin".format(uuid.uuid4())

    @classmethod
    def for_record_and_name(cls, record: SWORDDeposit, packaging_name: str):
        packaging_class = current_app.config["SWORD_ENDPOINTS"][record.pid.pid_type][
            "packaging_formats"
        ][packaging_name]
        return packaging_class(record)

    def shortcut_unpack(
        self, object_version: ObjectVersion
    ) -> Union[NotImplemented, Collection[str]]:
        """Override this to shortcut task-based unpacking"""
        return NotImplemented

    def unpack(self, object_version: ObjectVersion) -> Collection[str]:
        raise NotImplementedError  # pragma: nocover



from .bagit import SWORDBagItPackaging
from .base import Packaging
from .binary import BinaryPackaging
from .zip import SimpleZipPackaging

__all__ = [
    "BinaryPackaging",
    "Packaging",
    "SWORDBagItPackaging",
    "SimpleZipPackaging",
]



from typing import Dict

import pkg_resources
from invenio_deposit.config import DEPOSIT_REST_ENDPOINTS
from sword3common.constants import MetadataFormat
from sword3common.constants import PackagingFormat

from . import permissions
from .typing import SwordEndpointDefinition

SWORD_MAX_UPLOAD_SIZE = 1024 ** 3  # 1 GiB
SWORD_MAX_BY_REFERENCE_SIZE = 10 * 1024 ** 3  # 10 GiB

_PID = 'pid(depid,record_class="invenio_sword.api:SWORDDeposit")'

SWORD_ENDPOINTS: Dict[str, SwordEndpointDefinition] = {
    name: {
        "packaging_formats": {
            ep.name: ep.load()
            for ep in pkg_resources.iter_entry_points("invenio_sword.packaging")
        },
        "metadata_formats": {
            ep.name: ep.load()
            for ep in pkg_resources.iter_entry_points("invenio_sword.metadata")
        },
        "default_packaging_format": PackagingFormat.Binary,
        "default_metadata_format": MetadataFormat.Sword,
        "record_class": "invenio_sword.api:Deposit",
        "default_media_type": "application/ld+json",
        # Permissions
        "create_permission_factory_imp": permissions.check_has_write_scope,
        "read_permission_factory_imp": permissions.check_is_record_owner,
        "update_permission_factory_imp": permissions.check_has_write_scope_and_is_record_owner,
        "delete_permission_factory_imp": permissions.check_has_write_scope_and_is_record_owner,
        # Routes
        "service_document_route": "/sword/service-document",
        "item_route": "/sword/deposit/<{}:pid_value>".format(_PID),
        "metadata_route": "/sword/deposit/<{}:pid_value>/metadata".format(_PID),
        "fileset_route": "/sword/deposit/<{}:pid_value>/fileset".format(_PID),
        "file_route": "/sword/deposit/<{}:pid_value>/file/<path:key>".format(_PID),
        # Search
        "search_class": "invenio_deposit.search:DepositSearch",
        "indexer_class": None,
        "search_index": None,
        "search_type": None,
        # By-reference
        "dereference_policy": (
            lambda object_version, by_reference_file: by_reference_file.dereference
        ),
    }
    for name, options in DEPOSIT_REST_ENDPOINTS.items()
}


_SEGMENTED_UPLOAD_PID = (
    'pid(stagingid,record_class="invenio_sword.api:SegmentedUploadRecord")'
)

SWORD_STAGING_MAX_IDLE = 3600
SWORD_STAGING_PID_TYPE = "stagingid"
SWORD_STAGING_URL_ROUTE = "/sword/staging"
SWORD_TEMPORARY_URL_ROUTE = "/sword/staging/<uuid:temporary_id>"
SWORD_SEGMENTED_UPLOAD_CONTEXT = {
    "create_permission_factory": permissions.check_has_write_scope,
    "read_permission_factory": permissions.check_is_record_owner,
    "update_permission_factory": permissions.check_has_write_scope_and_is_record_owner,
    "delete_permission_factory": permissions.check_has_write_scope_and_is_record_owner,
}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.
"""Test the BagIt implementation."""
import os

from invenio_files_rest.models import ObjectVersion

from invenio_sword.api import SWORDDeposit
from invenio_sword.packaging import SimpleZipPackaging

fixtures_path = os.path.join(os.path.dirname(__file__), "fixtures")


def test_simple_zip(api, users, location):
    with api.test_request_context():
        record = SWORDDeposit.create({})
        with open(os.path.join(fixtures_path, "simple.zip"), "rb") as stream:
            object_version = ObjectVersion.create(
                bucket=record.bucket,
                key="deposit.zip",
                stream=stream,
                mimetype="application/zip",
            )

        SimpleZipPackaging(record).unpack(object_version)

        obj_1 = ObjectVersion.query.filter_by(
            bucket=record.bucket, key="example.svg"
        ).one()
        obj_2 = ObjectVersion.query.filter_by(
            bucket=record.bucket, key="hello.txt"
        ).one()

        assert obj_1.mimetype == "image/svg+xml"
        assert obj_2.mimetype == "text/plain"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.
"""Test the API."""
import re
from http import HTTPStatus

from flask_security import url_for_security
from invenio_files_rest.models import ObjectVersion

from invenio_sword.api import pid_resolver


def test_get_service_document(api):
    with api.test_client() as client:
        response = client.get("/sword/service-document")
        assert response.status_code == HTTPStatus.OK
        assert response.is_json


def test_deposit_empty(api, users, location):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post("/sword/service-document")
        assert response.status_code == HTTPStatus.CREATED
        match = re.match(
            "^http://localhost/sword/deposit/([^/]+)$", response.headers["Location"]
        )
        assert match is not None
        pid_value = match.group(1)
        _, record = pid_resolver.resolve(pid_value)
        assert dict(record) == {
            "metadata": {},
            "$schema": "http://localhost/schemas/deposits/deposit-v1.0.0.json",
            "_deposit": {
                "id": pid_value,
                "status": "published",
                "owners": [users[0]["id"]],
                "created_by": users[0]["id"],
            },
            "_bucket": record.bucket_id,
        }

        # POSTing with no metadata, by-reference, or request body should result in no files being created.
        assert ObjectVersion.query.filter_by(bucket=record.bucket).count() == 0


def test_metadata_deposit(api, users, location, metadata_document):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post(
            "/sword/service-document",
            data=metadata_document,
            headers={
                "Content-Disposition": "attachment; metadata=true",
                "Content-Type": "application/ld+json",
            },
        )
        assert response.status_code == HTTPStatus.CREATED
        match = re.match(
            "^http://localhost/sword/deposit/([^/]+)$", response.headers["Location"]
        )
        assert match is not None
        pid_value = match.group(1)
        _, record = pid_resolver.resolve(pid_value)
        assert dict(record) == {
            "metadata": {"title_statement": {"title": "The title"}},
            "swordMetadata": {
                "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                "@type": "Metadata",
                "dc:contributor": "A.N. Other",
                "dc:title": "The title",
                "dcterms:abstract": "This is my abstract",
            },
            "swordMetadataSourceFormat": "http://purl.org/net/sword/3.0/types/Metadata",
            "$schema": "http://localhost/schemas/deposits/deposit-v1.0.0.json",
            "_deposit": {
                "id": pid_value,
                "status": "published",
                "owners": [users[0]["id"]],
                "created_by": users[0]["id"],
            },
            "_bucket": record.bucket_id,
        }



from __future__ import annotations

import functools
import io
import json
import logging
import typing

import celery
from flask import url_for
from invenio_db import db
from sqlalchemy import true
from werkzeug.exceptions import Conflict
from werkzeug.http import parse_options_header

from invenio_deposit.api import Deposit
from invenio_deposit.api import has_status
from invenio_files_rest.models import ObjectVersion
from invenio_files_rest.models import ObjectVersionTag
from invenio_pidstore.resolver import Resolver
from invenio_records_files.api import FileObject, Record
from invenio_records_files.api import FilesIterator
from invenio_sword.enum import FileState
from invenio_sword.enum import ObjectTagKey
from invenio_sword.metadata import SWORDMetadata
from invenio_sword.typing import BytesReader
from sword3common.constants import DepositState
from sword3common.constants import Rel
from .metadata import Metadata
from .packaging import Packaging
from .schemas import ByReferenceFileDefinition
from .utils import TagManager

logger = logging.getLogger(__name__)


class SWORDFileObject(FileObject):
    def __init__(self, *args, pid, **kwargs):
        self.pid = pid
        return super().__init__(*args, **kwargs)

    @property
    def rest_file_url(self):
        return url_for(
            "invenio_sword.{}_file".format(self.pid.pid_type),
            pid_value=self.pid.pid_value,
            key=self.key,
            _external=True,
        )


class SWORDFilesIterator(FilesIterator):
    def _query(self):
        return ObjectVersion.query.outerjoin(ObjectVersionTag).filter(
            ObjectVersion.bucket == self.bucket,
            ObjectVersion.is_head == true(),
            ObjectVersion.file_id.isnot(None)
            | (
                (ObjectVersionTag.key == ObjectTagKey.ByReferenceNotDeleted.value)
                & (ObjectVersionTag.value == "true")
            ),
        )

    def __len__(self):
        return self._query().count()

    def __iter__(self):
        """Get iterator."""
        self._it = iter(self._query())
        return self


class SWORDDeposit(Deposit):
    @property
    def file_cls(self):
        return functools.partial(SWORDFileObject, pid=self.pid)

    files_iter_cls = SWORDFilesIterator

    def get_status_as_jsonld(self):
        editable = self["_deposit"].get("status") == "draft"

        return {
            "@id": self.sword_status_url,
            "@type": "Status",
            "metadata": {"@id": self.sword_metadata_url,},
            "fileSet": {"@id": self.sword_fileset_url,},
            "service": url_for(
                "invenio_sword.{}_service_document".format(self.pid.pid_type)
            ),
            "state": self.sword_states,
            "actions": {
                "getMetadata": True,
                "getFiles": True,
                "appendMetadata": editable,
                "appendFiles": editable,
                "replaceMetadata": editable,
                "replaceFiles": editable,
                "deleteMetadata": editable,
                "deleteFiles": editable,
                "deleteObject": editable,
            },
            "links": self.links,
        }

    @property
    def links(self):

        links = []
        for file in self.files or ():
            tags = TagManager(file)

            link = {
                "@id": file.rest_file_url,
                "contentType": file.obj.mimetype,
                "status": tags.get(ObjectTagKey.FileState, FileState.Ingested).value,
            }

            rel = set()
            if tags.get(ObjectTagKey.OriginalDeposit) == "true":
                rel.add(Rel.OriginalDeposit)
            if tags.get(ObjectTagKey.FileSetFile) == "true":
                rel.add(Rel.FileSetFile)
            derived_from = tags.get(ObjectTagKey.DerivedFrom)
            if derived_from:
                rel.add(Rel.DerivedResource)
                link["derivedFrom"] = url_for(
                    "invenio_sword.{}_file".format(self.pid.pid_type),
                    pid_value=self.pid.pid_value,
                    key=derived_from,
                    _external=True,
                )
            if ObjectTagKey.Packaging in tags:
                link["packaging"] = tags[ObjectTagKey.Packaging]
            if ObjectTagKey.MetadataFormat in tags:
                rel.add(Rel.FormattedMetadata)
                link["metadataFormat"] = tags[ObjectTagKey.MetadataFormat]

            link["rel"] = sorted(rel)

            links.append(link)

        return links

    @property
    def sword_states(self):
        states = []
        if self["_deposit"].get("status") == "draft":
            states.append(
                {
                    "@id": DepositState.InProgress,
                    "description": "the item is currently inProgress",
                }
            )
        elif self["_deposit"].get("status") == "published":
            states.append(
                {"@id": DepositState.Ingested, "description": "the item is ingested",}
            )
        return states

    @property
    def sword_status_url(self):
        return url_for(
            "invenio_sword.{}_item".format(self.pid.pid_type),
            pid_value=self.pid.pid_value,
            _external=True,
        )

    @property
    def sword_metadata_url(self):
        return url_for(
            "invenio_sword.{}_metadata".format(self.pid.pid_type),
            pid_value=self.pid.pid_value,
            _external=True,
        )

    @property
    def sword_fileset_url(self):
        return url_for(
            "invenio_sword.{}_fileset".format(self.pid.pid_type),
            pid_value=self.pid.pid_value,
            _external=True,
        )

    @property
    def metadata_key_prefix(self):
        return ".metadata-{}/".format(self.pid.pid_value)

    @property
    def original_deposit_key_prefix(self):
        return ".original-deposit-{}/".format(self.pid.pid_value)

    @has_status(status="draft")
    def ingest_file(
        self,
        stream: typing.Optional[BytesReader],
        packaging_name: str,
        content_disposition: str = None,
        content_type: str = None,
        replace=True,
    ):
        from . import tasks

        if stream:
            packaging = Packaging.for_record_and_name(self, packaging_name)

            content_disposition, content_disposition_options = parse_options_header(
                content_disposition or ""
            )
            content_type, _ = parse_options_header(content_type or "")
            filename = packaging.get_original_deposit_filename(
                content_disposition_options.get("filename"), content_type
            )
            object_version = ObjectVersion.create(
                bucket=self.bucket, key=filename, stream=stream
            )
            TagManager(object_version).update(
                {
                    ObjectTagKey.Packaging: packaging_name,
                    ObjectTagKey.OriginalDeposit: "true",
                }
            )
            db.session.refresh(object_version)
            task = self.unpack_object(object_version)
            if replace:
                task |= tasks.delete_old_objects.s(bucket_id=self.bucket_id)
            task.delay()
        elif replace:
            # We can do this synchronously, because it'll be quick
            tasks.delete_old_objects(bucket_id=self.bucket_id)

    # @has_status(status="draft")
    def set_metadata(
        self,
        source: typing.Optional[typing.Union[BytesReader, dict]],
        metadata_class: typing.Type[Metadata],
        content_type: str = None,
        derived_from: str = None,
        replace: bool = True,
    ) -> typing.Optional[Metadata]:
        if isinstance(source, dict):
            source = io.BytesIO(json.dumps(source).encode("utf-8"))

        if not content_type:
            content_type = metadata_class.content_type

        existing_metadata_object = (
            ObjectVersion.query.join(ObjectVersion.tags)
            .filter(
                ObjectVersion.is_head == true(),
                ObjectVersion.file_id.isnot(None),
                ObjectVersion.bucket == self.bucket,
                ObjectVersionTag.key == ObjectTagKey.MetadataFormat.value,
                ObjectVersionTag.value == metadata_class.metadata_format,
            )
            .first()
        )

        if source is None:
            if replace and existing_metadata_object:
                ObjectVersion.delete(
                    bucket=existing_metadata_object.bucket,
                    key=existing_metadata_object.key,
                )

            if replace and (
                self.get("swordMetadataSourceFormat") == metadata_class.metadata_format
            ):
                self.pop("swordMetadata", None)
                self.pop("swordMetadataSourceFormat", None)

            return None
        else:
            content_type, content_type_options = parse_options_header(content_type)

            encoding = content_type_options.get("charset")
            if isinstance(encoding, str):
                metadata = metadata_class.from_document(
                    source, content_type=content_type, encoding=encoding,
                )
            else:
                metadata = metadata_class.from_document(
                    source, content_type=content_type,
                )

            if existing_metadata_object and not replace:
                with existing_metadata_object.file.storage().open() as existing_metadata_f:
                    existing_metadata = metadata_class.from_document(
                        existing_metadata_f, content_type=metadata_class.content_type,
                    )
                try:
                    metadata = existing_metadata + metadata
                except TypeError:
                    raise Conflict(
                        "Existing or new metadata is of wrong type for appending. Reconcile client-side and PUT instead"
                    )

            metadata_filename = self.metadata_key_prefix + metadata_class.filename

            if (
                isinstance(metadata, SWORDMetadata)
                or "swordMetadata" not in self
                or (
                    not isinstance(metadata, SWORDMetadata)
                    and self["swordMetadataSourceFormat"]
                    == metadata_class.metadata_format
                )
            ):
                metadata.update_record_metadata(self)
                self["swordMetadata"] = metadata.to_sword_metadata()
                self["swordMetadataSourceFormat"] = metadata_class.metadata_format

            object_version = ObjectVersion.create(
                bucket=self.bucket,
                key=metadata_filename,
                stream=io.BytesIO(bytes(metadata)),
            )

            tags = TagManager(object_version)
            tags[ObjectTagKey.MetadataFormat] = metadata_class.metadata_format
            if derived_from:
                tags[ObjectTagKey.DerivedFrom] = derived_from

            return metadata

    def set_by_reference_files(
        self,
        by_reference_files: typing.Collection[ByReferenceFileDefinition],
        dereference_policy,
        request_url,
        replace=True,
    ):
        from . import tasks

        task_group = []

        for by_reference_file in by_reference_files:
            content_disposition, content_disposition_options = parse_options_header(
                by_reference_file.content_disposition,
            )
            content_type, _ = parse_options_header(by_reference_file.content_type)
            filename = content_disposition_options["filename"]
            object_version = ObjectVersion.create(
                self.bucket, filename, mimetype=content_type
            )
            tags = TagManager(object_version)
            tags.update(
                {
                    ObjectTagKey.FileState: FileState.Pending,
                    ObjectTagKey.OriginalDeposit: "true",
                    ObjectTagKey.Packaging: by_reference_file.packaging,
                    ObjectTagKey.ByReferenceDereference: (
                        "true" if by_reference_file.dereference else "false"
                    ),
                    ObjectTagKey.ByReferenceNotDeleted: "true",
                }
            )
            if by_reference_file.url:
                tags[ObjectTagKey.ByReferenceURL] = by_reference_file.url
            elif by_reference_file.temporary_id:
                tags[ObjectTagKey.ByReferenceTemporaryID] = str(
                    by_reference_file.temporary_id
                )

            if by_reference_file.ttl:
                tags[ObjectTagKey.ByReferenceTTL] = by_reference_file.ttl.isoformat()
            if by_reference_file.content_length:
                tags[ObjectTagKey.ByReferenceContentLength] = str(
                    by_reference_file.content_length
                )

            if dereference_policy(object_version, by_reference_file):
                # Need to refresh so that self.dereference_object can see the tags
                db.session.refresh(object_version)
                task_group.append(self.dereference_object(object_version))

        if task_group:
            # Don't need to group if there's only one task, which avoids a chord
            task = celery.group(task_group) if len(task_group) > 1 else task_group[0]
            if replace:
                task |= tasks.delete_old_objects.s(bucket_id=self.bucket_id)
            task.delay()
        elif replace:
            tasks.delete_old_objects(bucket_id=self.bucket_id)

    def dereference_object(self, object_version: ObjectVersion):
        """Queues a task to dereference an object"""
        from . import tasks

        tags = TagManager(object_version)
        assert tags[ObjectTagKey.Packaging]
        assert tags.get(ObjectTagKey.ByReferenceURL) or tags.get(
            ObjectTagKey.ByReferenceTemporaryID
        )
        assert object_version.is_head

        tags[ObjectTagKey.FileState] = FileState.Pending

        task_signature = tasks.dereference_object.s(
            str(self.id), str(object_version.version_id)
        )
        task_signature |= tasks.unpack_object.si(
            str(self.id), str(object_version.version_id)
        )
        return task_signature

    def unpack_object(self, object_version: ObjectVersion):
        """Queues a task to unpack an object"""
        from . import tasks

        tags = TagManager(object_version)
        assert tags[ObjectTagKey.Packaging]
        assert object_version.is_head

        tags[ObjectTagKey.FileState] = FileState.Unpacking

        return tasks.unpack_object.s(str(self.id), str(object_version.version_id))


class SegmentedUploadRecord(Record):
    pass


pid_resolver = Resolver(
    pid_type="depid",
    object_type="rec",
    getter=functools.partial(SWORDDeposit.get_record, with_deleted=True),
)



"""
Segmented file upload support

See <https://swordapp.github.io/swordv3/swordv3.html#9.7.>.
"""
import datetime

import math
from http import HTTPStatus

from flask_login import current_user
from invenio_db import db
from sqlalchemy.orm.exc import NoResultFound

import sword3common.exceptions
from flask import request, url_for, Response
from werkzeug.http import parse_options_header

from invenio_files_rest.errors import MultipartMissingParts, UnexpectedFileSizeError
from invenio_files_rest.models import MultipartObject, Part
from invenio_records_rest.views import need_record_permission
from invenio_sword.api import SegmentedUploadRecord

from invenio_sword.schemas import SegmentInitSchema, SegmentUploadSchema
from invenio_sword.views import SWORDDepositView


__all__ = ["StagingURLView", "TemporaryURLView"]


class SegmentedUploadView(SWORDDepositView):
    record_class = SegmentedUploadRecord


class StagingURLView(SegmentedUploadView):
    """POST to this to start a segmented file upload and create a temporary URL"""

    view_name = "staging_url"

    @need_record_permission("create_permission_factory")
    def post(self, **kwargs):
        content_disposition, content_disposition_options = parse_options_header(
            request.headers.get("Content-Disposition", "")
        )
        if content_disposition != "segment-init":
            raise sword3common.exceptions.BadRequest(
                "Content-Disposition must be 'segment-init'"
            )
        parsed_content_disposition_options = SegmentInitSchema().load(
            content_disposition_options
        )

        record: SegmentedUploadRecord = SegmentedUploadRecord.create(
            {"_segmentedUpload": {"owners": [int(current_user.get_id())],}}
        )
        MultipartObject.create(
            bucket=record.bucket,
            key=str(record.id),
            size=parsed_content_disposition_options["size"],
            chunk_size=parsed_content_disposition_options["segment_size"],
        )

        response = Response(status=HTTPStatus.CREATED)
        response.headers["Location"] = url_for(
            "invenio_sword.temporary_url", temporary_id=record.id, _external=True,
        )
        db.session.commit()
        return response


class TemporaryURLView(SegmentedUploadView):
    """Upload file segments here"""

    view_name = "temporary_url"
    record_class = SegmentedUploadRecord

    def dispatch_request(self, temporary_id, *args, **kwargs):
        try:
            record = SegmentedUploadRecord.get_record(temporary_id)
        except NoResultFound as e:
            raise sword3common.exceptions.NotFound from e
        if int(current_user.get_id()) not in record["_segmentedUpload"]["owners"]:
            raise sword3common.exceptions.SwordException.for_status_code_and_name(
                HTTPStatus.FORBIDDEN, None
            )
        multipart_object: MultipartObject = MultipartObject.query.filter_by(
            bucket_id=record.bucket_id,
        ).one()
        return super().dispatch_request(
            record=record, multipart_object=multipart_object, *args, **kwargs
        )

    def get(self, *, record: SegmentedUploadRecord, multipart_object: MultipartObject):
        part_numbers = set(
            db.session.query(Part.part_number).filter_by(
                upload_id=multipart_object.upload_id
            )
        )

        segment_count = math.ceil(multipart_object.size / multipart_object.chunk_size)

        return {
            "@type": "Temporary",
            "segments": {
                "received": sorted(part_numbers),
                "expecting": [
                    i for i in range(1, segment_count + 1) if i not in part_numbers
                ],
                "size": multipart_object.size,
                "segment_size": multipart_object.chunk_size,
            },
        }

    def post(self, *, record: SegmentedUploadRecord, multipart_object: MultipartObject):
        content_disposition, content_disposition_options = parse_options_header(
            request.headers.get("Content-Disposition", "")
        )
        if content_disposition != "segment":
            raise sword3common.exceptions.BadRequest(
                "Content-Disposition must be 'segment'"
            )
        parsed_content_disposition_options = SegmentUploadSchema(
            segment_count=multipart_object.last_part_number + 1
        ).load(content_disposition_options)

        try:
            Part.create(
                multipart_object,
                # SWORD segment_numbers are indexed from 1, whereas invenio part numbers are indexed from 0
                parsed_content_disposition_options["segment_number"] - 1,
                request.stream,
            )
        except UnexpectedFileSizeError as e:
            raise sword3common.exceptions.InvalidSegmentSize(e.description) from e

        # Attempt to mark it complete, as a segmented upload is complete as soon as the last part is uploaded, and
        # this method does that check for us
        try:
            multipart_object.complete()
        except MultipartMissingParts:
            pass

        multipart_object.updated = datetime.datetime.utcnow()
        db.session.add(multipart_object)

        db.session.commit()
        return Response(status=HTTPStatus.NO_CONTENT)

    def delete(
        self, *, record: SegmentedUploadRecord, multipart_object: MultipartObject
    ):
        multipart_object.delete()
        db.session.commit()
        return Response(status=HTTPStatus.NO_CONTENT)



import pytest
from invenio_db import db
from invenio_files_rest.models import Bucket
from invenio_files_rest.models import ObjectVersion
from invenio_files_rest.models import ObjectVersionTag

from invenio_sword.enum import ObjectTagKey
from invenio_sword.utils import TagManager


def test_tag_manager_setitem(api, users, location, es):
    value = "http://example.org/"

    with api.test_request_context():
        bucket = Bucket.create()
        object_version = ObjectVersion.create(bucket=bucket, key="hello")
        tags = TagManager(object_version)

        assert (
            ObjectVersionTag.query.filter_by(object_version=object_version).count() == 0
        )
        tags[ObjectTagKey.Packaging] = value

        # It's updated the TagManager
        assert tags[ObjectTagKey.Packaging] == value

        # We've created a tag
        tag = ObjectVersionTag.query.filter_by(object_version=object_version).one()
        assert tag.key == ObjectTagKey.Packaging.value
        assert tag.value == value


def test_tag_manager_delitem(api, users, location, es):
    value = "http://example.org/"

    with api.test_request_context():
        bucket = Bucket.create()
        object_version = ObjectVersion.create(bucket=bucket, key="hello")
        ObjectVersionTag.create(
            object_version=object_version, key=ObjectTagKey.Packaging.value, value=value
        )
        tags = TagManager(object_version)

        assert tags == {ObjectTagKey.Packaging: value}
        assert tags[ObjectTagKey.Packaging] == value

        del tags[ObjectTagKey.Packaging]

        assert tags == {}
        with pytest.raises(KeyError):
            _ = tags[ObjectTagKey.Packaging]

        # We've deleted the database object
        assert (
            ObjectVersionTag.query.filter_by(object_version=object_version).count() == 0
        )


@pytest.mark.parametrize("update_style", ["dict", "kwargs"])
def test_tag_manager_update(api, users, location, es, update_style):
    with api.test_request_context():
        bucket = Bucket.create()
        object_version = ObjectVersion.create(bucket=bucket, key="hello")
        ObjectVersionTag.create(
            object_version=object_version,
            key=ObjectTagKey.Packaging.value,
            value="old-packaging",
        )
        ObjectVersionTag.create(
            object_version=object_version,
            key=ObjectTagKey.MetadataFormat.value,
            value="old-metadata",
        )
        tags = TagManager(object_version)

        assert (
            ObjectVersionTag.query.filter_by(object_version=object_version).count() == 2
        )

        assert tags == {
            ObjectTagKey.Packaging: "old-packaging",
            ObjectTagKey.MetadataFormat: "old-metadata",
        }

        if update_style == "dict":
            tags.update(
                {
                    ObjectTagKey.MetadataFormat: "new-metadata",
                    ObjectTagKey.DerivedFrom: "new-derived-from",
                }
            )
        elif update_style == "kwargs":
            tags.update(
                **{
                    ObjectTagKey.MetadataFormat.value: "new-metadata",
                    ObjectTagKey.DerivedFrom.value: "new-derived-from",
                }
            )

        assert tags == {
            ObjectTagKey.Packaging: "old-packaging",
            ObjectTagKey.MetadataFormat: "new-metadata",
            ObjectTagKey.DerivedFrom: "new-derived-from",
        }

        assert (
            ObjectVersionTag.query.filter_by(object_version=object_version).count() == 3
        )

        db.session.refresh(object_version)
        assert object_version.get_tags() == {
            ObjectTagKey.Packaging.value: "old-packaging",
            ObjectTagKey.MetadataFormat.value: "new-metadata",
            ObjectTagKey.DerivedFrom.value: "new-derived-from",
        }



from http import HTTPStatus

from flask import current_app, url_for
from flask import request
from flask import Response
from invenio_records_rest.views import need_record_permission
from invenio_records_rest.views import verify_record_permission

from . import SWORDDepositView

__all__ = ["ServiceDocumentView"]


class ServiceDocumentView(SWORDDepositView):
    view_name = "{}_service_document"

    def get(self):
        """Retrieve the service document

        :see also: https://swordapp.github.io/swordv3/swordv3.html#9.2.
        """
        return {
            # Preamble
            "@type": "ServiceDocument",
            "dc:title": current_app.config["THEME_SITENAME"],
            "root": request.url,
            "acceptDeposits": True,
            "version": "http://purl.org/net/sword/3.0",
            # Capabilities
            "byReferenceDeposit": True,
            "onBehalfOf": False,
            # Size limits (not currently enforced)
            "maxUploadSize": current_app.config["SWORD_MAX_UPLOAD_SIZE"],
            "maxByReferenceSize": current_app.config["SWORD_MAX_BY_REFERENCE_SIZE"],
            # Accepted formats
            "acceptArchiveFormat": ["application/zip"],
            "acceptPackaging": sorted(self.endpoint_options["packaging_formats"]),
            "acceptMetadata": sorted(self.endpoint_options["metadata_formats"]),
            # Segmented uploads
            "staging": url_for("invenio_sword.staging_url"),
            "stagingMaxIdle": current_app.config["SWORD_STAGING_MAX_IDLE"],
            "maxAssembledSize": min(
                current_app.config["FILES_REST_MULTIPART_MAX_PARTS"]
                * current_app.config["FILES_REST_MULTIPART_CHUNKSIZE_MAX"],
                current_app.config["SWORD_MAX_BY_REFERENCE_SIZE"],
            ),
            "maxSegments": current_app.config["FILES_REST_MULTIPART_MAX_PARTS"],
        }

    @need_record_permission("create_permission_factory")
    def post(self, **kwargs):
        """Initiate a SWORD deposit"""
        record = self.create_deposit()

        # Check permissions
        permission_factory = self.create_permission_factory
        if permission_factory:
            verify_record_permission(permission_factory, record)

        self.update_deposit(record, replace=False)

        response = self.make_response(record.get_status_as_jsonld())  # type: Response
        response.status_code = HTTPStatus.CREATED
        response.headers["Location"] = record.sword_status_url
        return response



import io
import secrets
from typing import Type

import pytest
from invenio_files_rest.models import ObjectVersion

from invenio_sword.api import SWORDDeposit
from invenio_sword.packaging import BinaryPackaging
from invenio_sword.packaging import Packaging
from invenio_sword.packaging import SimpleZipPackaging
from invenio_sword.packaging import SWORDBagItPackaging


def test_get_original_deposit_filename(api, es, location):
    with api.test_request_context():
        record: SWORDDeposit = SWORDDeposit.create({})
        packaging = BinaryPackaging(record)

        filename = secrets.token_hex(16)
        assert (
            packaging.get_original_deposit_filename(filename, media_type="text/plain")
            == filename
        )

        assert packaging.get_original_deposit_filename(
            media_type="text/plain"
        ).endswith(".txt")

        assert packaging.get_original_deposit_filename().endswith(".bin")


@pytest.mark.parametrize("packaging_cls", [SimpleZipPackaging, SWORDBagItPackaging])
def test_non_binary_doesnt_shortcut_unpack(
    api, location, es, packaging_cls: Type[Packaging]
):
    with api.test_request_context():
        record = SWORDDeposit.create({})
        object_version = ObjectVersion.create(
            bucket=record.bucket, key="some-file.txt", stream=io.BytesIO(b"data")
        )
        packaging = packaging_cls(record)
        assert packaging.shortcut_unpack(object_version) == NotImplemented



import logging
import urllib.request
import uuid
from typing import Iterable
from typing import Sequence
from typing import Union

import celery
from celery.result import AsyncResult
from invenio_db import db
from sqlalchemy import true

from invenio_files_rest.models import MultipartObject, ObjectVersion
from invenio_files_rest.models import ObjectVersionTag
from invenio_sword.api import SWORDDeposit, SegmentedUploadRecord
from invenio_sword.enum import FileState
from invenio_sword.enum import ObjectTagKey
from invenio_sword.packaging import Packaging
from invenio_sword.utils import TagManager

logger = logging.getLogger(__name__)


@celery.shared_task
def dereference_object(record_id, version_id):
    object_version = ObjectVersion.query.filter(
        ObjectVersion.version_id == version_id
    ).one()
    if not object_version.is_head:
        logger.info(
            "Not fetching by-reference file (%s) because a newer version of the object now exists",
            object_version,
        )
        return [object_version.key]

    if object_version.file_id:
        logger.warning("File has already been dereferenced (%s)", object_version)
        return [object_version.key]

    tags = TagManager(object_version)

    try:
        if ObjectTagKey.ByReferenceURL in tags:
            tags[ObjectTagKey.FileState] = FileState.Downloading
            url = tags[ObjectTagKey.ByReferenceURL]
            response = urllib.request.urlopen(url)
            object_version.set_contents(response)
            pass
        elif ObjectTagKey.ByReferenceTemporaryID in tags:
            tags[ObjectTagKey.FileState] = FileState.Downloading
            temporary_id = uuid.UUID(tags[ObjectTagKey.ByReferenceTemporaryID])
            segmented_upload_record = SegmentedUploadRecord.get_record(temporary_id)
            multipart_object: MultipartObject = MultipartObject.query.filter_by(
                bucket_id=segmented_upload_record.bucket_id
            ).one()
            if not multipart_object.completed:
                raise ValueError(
                    "Segmented upload must be completed before it can be ingested by-reference"
                )
            object_version.file_id = multipart_object.file_id
        else:
            raise ValueError(
                "Missing URL for by-reference file {!r}".format(object_version)
            )
        tags[ObjectTagKey.FileState] = FileState.Pending
        del tags[ObjectTagKey.ByReferenceNotDeleted]
        return [object_version.key]
    except Exception:
        logger.exception("Error retrieving by-reference file")
        tags[ObjectTagKey.FileState] = FileState.Error
        raise
    finally:
        db.session.commit()


@celery.shared_task
def unpack_object(record_id, version_id):
    record = SWORDDeposit.get_record(record_id)

    object_version: ObjectVersion = ObjectVersion.query.filter(
        ObjectVersion.version_id == version_id
    ).one()
    if not object_version.is_head:
        logger.info(
            "Not fetching by-reference file (%s) because a newer version of the object now exists",
            object_version,
        )
        return

    tags = TagManager(object_version)

    try:
        packaging = Packaging.for_record_and_name(record, tags[ObjectTagKey.Packaging])
        keys = packaging.unpack(object_version)
        tags[ObjectTagKey.FileState] = FileState.Ingested
        return [object_version.key] + list(keys)
    except Exception:
        logger.exception(
            "Failed to unpack %s:%s", object_version.bucket_id, object_version.key
        )
        tags[ObjectTagKey.FileState] = FileState.Error
        raise
    finally:
        db.session.commit()


@celery.shared_task
def delete_old_objects(
    ignore_keys: Union[Sequence[AsyncResult], Iterable[str]] = (), *, bucket_id: str
):
    for object_version in ObjectVersion.query.join(ObjectVersionTag).filter(
        ObjectVersion.bucket_id == bucket_id,
        ObjectVersion.key.notin_(ignore_keys),
        ObjectVersion.is_head == true(),
        ObjectVersionTag.key.in_(
            [
                ObjectTagKey.FileSetFile.value,
                ObjectTagKey.DerivedFrom.value,
                ObjectTagKey.OriginalDeposit.value,
            ]
        ),
    ):
        if object_version.file_id:
            # Delete any extant files
            ObjectVersion.delete(object_version.bucket, object_version.key)
        else:
            # Delete any tags that say that an ObjectVersion is a yet-to-be-dereferenced file
            ObjectVersionTag.query.filter(
                ObjectVersionTag.version_id == object_version.version_id,
                ObjectVersionTag.key == ObjectTagKey.ByReferenceNotDeleted.value,
                ObjectVersionTag.value == "true",
            ).delete()

    db.session.commit()



import io
import json
import os
from http import HTTPStatus

from flask import url_for
from flask_security import url_for_security
from invenio_db import db
from invenio_files_rest.models import ObjectVersion
from invenio_files_rest.models import ObjectVersionTag
from sqlalchemy import null
from sqlalchemy import true

from invenio_sword.api import SWORDDeposit
from invenio_sword.enum import ObjectTagKey
from invenio_sword.packaging import SWORDBagItPackaging


def test_get_fileset_url(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        db.session.commit()

        response = client.get(
            url_for("invenio_sword.depid_fileset", pid_value=record.pid.pid_value)
        )
        assert response.status_code == HTTPStatus.METHOD_NOT_ALLOWED


def test_put_fileset_url(api, users, location, es, task_delay):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        object_version = ObjectVersion.create(
            record.bucket,
            key="old-file.txt",
            stream=io.BytesIO(b"hello"),
            mimetype="text/plain",
        )
        ObjectVersionTag.create(
            object_version=object_version,
            key=ObjectTagKey.FileSetFile.value,
            value="true",
        )
        db.session.commit()

        response = client.put(
            url_for("invenio_sword.depid_fileset", pid_value=record.pid.pid_value),
            data=b"hello again",
            headers={
                "Content-Disposition": "attachment; filename=new-file.txt",
                "Content-Type": "text/plain",
            },
        )
        assert response.status_code == HTTPStatus.NO_CONTENT

        assert task_delay.call_count == 1
        task_self = task_delay.call_args[0][0]
        task_self.apply()

        # Check original ObjectVersion is marked deleted
        original_object_versions = list(
            ObjectVersion.query.filter_by(
                bucket=record.bucket, key="old-file.txt"
            ).order_by("created")
        )
        assert len(original_object_versions) == 2
        assert not original_object_versions[0].is_head
        assert original_object_versions[1].is_head
        assert original_object_versions[1].file is None

        # Check new object has been created
        new_object_version = ObjectVersion.query.filter_by(
            bucket=record.bucket, key="new-file.txt"
        ).one()
        assert new_object_version.is_head


def test_post_fileset_url(api, users, location, es, task_delay):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        ObjectVersion.create(
            record.bucket,
            key="old-file.txt",
            stream=io.BytesIO(b"hello"),
            mimetype="text/plain",
        )
        db.session.commit()

        response = client.post(
            url_for("invenio_sword.depid_fileset", pid_value=record.pid.pid_value),
            data=b"hello again",
            headers={
                "Content-Disposition": "attachment; filename=new-file.txt",
                "Content-Type": "text/plain",
            },
        )
        assert response.status_code == HTTPStatus.NO_CONTENT

        # Check original ObjectVersion is still there
        original_object_versions = list(
            ObjectVersion.query.filter_by(
                bucket=record.bucket, key="old-file.txt"
            ).order_by("created")
        )
        assert len(original_object_versions) == 1
        assert original_object_versions[0].is_head

        # Check new object has been created
        new_object_version = ObjectVersion.query.filter_by(
            bucket=record.bucket, key="new-file.txt"
        ).one()
        assert new_object_version.is_head


def test_delete_fileset(
    api, users, location, es, fixtures_path, test_metadata_format, task_delay
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        # Create a deposit, initially with a metadata deposit
        original_response = client.post(
            url_for("invenio_sword.depid_service_document"),
            data=json.dumps(
                {
                    "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                    "title": "A title",
                }
            ),
            headers={
                "Content-Disposition": "attachment; metadata=true",
                "In-Progress": "true",
            },
        )

        assert ObjectVersion.query.count() == 1

        # Add some extra metadata
        response = client.post(
            original_response.json["metadata"]["@id"],
            data=b"some metadata",
            headers={"Metadata-Format": test_metadata_format},
        )
        assert response.status_code == HTTPStatus.NO_CONTENT

        assert ObjectVersion.query.count() == 2

        with open(os.path.join(fixtures_path, "bagit.zip"), "rb") as f:
            response = client.post(
                original_response.headers["Location"],
                data=f,
                headers={
                    "Packaging": SWORDBagItPackaging.packaging_name,
                    "Content-Type": "application/zip",
                    "In-Progress": "true",
                },
            )
            assert response.status_code == HTTPStatus.OK

        assert task_delay.call_count == 1
        task_self = task_delay.call_args[0][0]
        task_self.apply()

        # One test metadata, one old SWORD metadata, one new SWORD metadata, one original deposit, and two files
        assert ObjectVersion.query.count() == 6
        # One test metadata, one new SWORD metadata, one original deposit, and two files
        assert ObjectVersion.query.filter(ObjectVersion.is_head == true()).count() == 5

        # Now let's delete the fileset. This should ensure that there is only one extant file, as the SWORD metadata and
        # original deposit were deposited as part of a fileset, leaving only the test dataset
        response = client.delete(original_response.json["fileSet"]["@id"])
        assert response.status_code == HTTPStatus.NO_CONTENT

        # All four previously extent files now have file_id=NULL versions
        assert ObjectVersion.query.count() == 10
        assert (
            ObjectVersion.query.filter(
                ObjectVersion.is_head == true(), ObjectVersion.file_id != null()
            ).count()
            == 1
        )



import typing

import sword3common.constants
import sword3common.exceptions
from flask import current_app
from flask import request
from invenio_db import db
from invenio_rest import ContentNegotiatedMethodView
from werkzeug.exceptions import Conflict
from werkzeug.http import parse_options_header
from werkzeug.utils import cached_property

from ..api import SWORDDeposit
from ..metadata import Metadata
from ..schemas import ByReferenceSchema
from ..typing import BytesReader

__all__ = ["SWORDDepositView"]


class SWORDDepositView(ContentNegotiatedMethodView):
    """
    Base class for all SWORD views

    The properties and methods defined on this class are used by subclasses to SWORD operations with common
    implementations across different SWORD views.
    """

    view_name: str
    record_class: typing.Type[SWORDDeposit]
    pid_type: str

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Constructor."""
        super(SWORDDepositView, self).__init__(
            serializers,
            default_media_type=ctx.get("default_media_type"),
            *args,
            **kwargs,
        )
        for key, value in ctx.items():
            setattr(self, key, value)

    @cached_property
    def endpoint_options(self) -> typing.Dict[str, typing.Any]:
        """Configuration endpoints for this view's SWORD endpoint"""
        return current_app.config["SWORD_ENDPOINTS"][self.pid_type]

    @cached_property
    def metadata_format(self) -> str:
        """The ``Metadata-Format`` header, or its default per config"""
        return request.headers.get(
            "Metadata-Format", self.endpoint_options["default_metadata_format"]
        )

    @cached_property
    def metadata_class(self) -> typing.Type[Metadata]:
        """The Metadata subclass associated with the request.

        :raises NotImplemented: if the ``Metadata-Format`` is not supported
        """
        try:
            return self.endpoint_options["metadata_formats"][self.metadata_format]
        except KeyError as e:
            raise sword3common.exceptions.MetadataFormatNotAcceptable from e

    @cached_property
    def packaging_name(self) -> str:
        """The Packaging subclass associated with the request.

        :raises NotImplemented: if the ``Packaging`` is not supported
        """
        packaging = request.headers.get(
            "Packaging", self.endpoint_options["default_packaging_format"]
        )
        if packaging not in self.endpoint_options["packaging_formats"]:
            raise sword3common.exceptions.PackagingFormatNotAcceptable
        return packaging

    @cached_property
    def in_progress(self) -> bool:
        """Whether the request declares that the deposit is still in progress, via the ``In-Progress`` header"""
        return request.headers.get("In-Progress") == "true"

    def update_deposit_status(self, record: SWORDDeposit) -> None:
        """Updates a deposit status from the In-Progress header

        """
        if self.in_progress:
            if record["_deposit"]["status"] == "published":
                raise Conflict(
                    "Deposit has status {}; cannot subsequently be made In-Progress".format(
                        record["_deposit"]["status"]
                    )
                )
        elif record["_deposit"]["status"] == "draft":
            record["_deposit"]["status"] = "published"

    def create_deposit(self) -> SWORDDeposit:
        """Create an empty deposit record"""
        return SWORDDeposit.create({"metadata": {}})

    def update_deposit(self, record: SWORDDeposit, replace: bool = True):
        """Update a deposit according to the request data

        :param replace: If ``True``, all previous data on the deposit is removed. If ``False``, the request augments
            data already provided.
        """

        content_disposition, content_disposition_options = parse_options_header(
            request.headers.get("Content-Disposition", "")
        )

        metadata_deposit = content_disposition_options.get("metadata") == "true"
        by_reference_deposit = content_disposition_options.get("by-reference") == "true"

        if metadata_deposit:
            if by_reference_deposit:  # pragma: nocover
                record.set_metadata(
                    request.json["metadata"], self.metadata_class, replace=replace,
                )
            else:
                record.set_metadata(
                    request.stream,
                    self.metadata_class,
                    request.content_type,
                    replace=replace,
                )
        elif replace:
            record.set_metadata(None, self.metadata_class, replace=replace)

        if by_reference_deposit:
            by_reference_schema = ByReferenceSchema(
                context={"url_adapter": current_app.create_url_adapter(request)},
            )
            if metadata_deposit:
                by_reference = by_reference_schema.load(request.json["by-reference"])
            else:
                by_reference = by_reference_schema.load(request.json)
            record.set_by_reference_files(
                by_reference["files"],
                dereference_policy=self.endpoint_options["dereference_policy"],
                request_url=request.url,
                replace=replace,
            )
        elif replace:
            record.set_by_reference_files(
                [],
                dereference_policy=self.endpoint_options["dereference_policy"],
                request_url=request.url,
                replace=replace,
            )

        if not (metadata_deposit or by_reference_deposit) and (
            request.content_type or request.content_length
        ):
            self.ingest_file(record, request.stream, replace=replace)
        elif replace:
            self.ingest_file(record, None, replace=replace)

        self.update_deposit_status(record)

        record.commit()
        db.session.commit()

    def ingest_file(
        self, record: SWORDDeposit, stream: typing.Optional[BytesReader], replace=True
    ):
        """
        Sets or adds to a deposit fileset using a bytestream and request headers

        This method wraps ``record.set_fileset_from_stream()`` with appropriate arguments.

        :param record: The SWORDDeposit record to modify
        :param stream: A bytestream of the file or package to be deposited
        :param replace: Whether to replace or add to the deposit
        :return: an IngestResult
        """
        return record.ingest_file(
            stream if (request.content_type or request.content_length) else None,
            packaging_name=self.packaging_name,
            content_disposition=request.headers.get("Content-Disposition"),
            content_type=request.content_type,
            replace=replace,
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.
"""Test the BagIt implementation."""
import os
from http import HTTPStatus

import pytest
from flask import url_for
from flask_security import url_for_security
from invenio_files_rest.models import Bucket
from invenio_files_rest.models import ObjectVersion
from invenio_records.models import RecordMetadata
from sword3common.constants import PackagingFormat
from sword3common.exceptions import ContentMalformed
from sword3common.exceptions import ContentTypeNotAcceptable
from sword3common.exceptions import ValidationFailed

from invenio_sword.api import SWORDDeposit
from invenio_sword.packaging import Packaging


def test_post_service_document_with_bagit_bag(
    api, users, location, fixtures_path, task_delay
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        with open(os.path.join(fixtures_path, "bagit.zip"), "rb") as f:
            response = client.post(
                url_for("invenio_sword.depid_service_document"),
                input_stream=f,
                headers={
                    "Content-Type": "application/zip",
                    "Packaging": "http://purl.org/net/sword/3.0/package/SWORDBagIt",
                },
            )

        assert response.status_code == HTTPStatus.CREATED

        # Check that we attempted to queue a task
        assert task_delay.call_count == 1
        task_self = task_delay.call_args[0][0]

        task_self.apply()

        # db.session.refresh(record)

        bucket = Bucket.query.one()
        obj_1 = ObjectVersion.query.filter_by(bucket=bucket, key="example.svg").one()
        obj_2 = ObjectVersion.query.filter_by(bucket=bucket, key="hello.txt").one()

        assert obj_1.mimetype == "image/svg+xml"
        assert obj_2.mimetype == "text/plain"

        record_metadata = RecordMetadata.query.one()
        assert record_metadata.json == {
            "$schema": "http://localhost/schemas/deposits/deposit-v1.0.0.json",
            "_bucket": str(bucket.id),
            "_deposit": {
                "created_by": 1,
                # A bit of a cheat, but we know this is the deposit pid_value, and it's maintained by invenio-deposit,
                # not the code that we are testing.
                "id": record_metadata.json["_deposit"]["id"],
                "owners": [1],
                "status": "published",
            },
            "bagitInfo": {
                "Bag-Software-Agent": "bagit.py v1.7.0 "
                "<https://github.com/LibraryOfCongress/bagit-python>",
                "Bagging-Date": "2020-01-07",
                "Payload-Oxum": "473.2",
            },
            "metadata": {"title_statement": {"title": "The title"}},
            "swordMetadata": {
                "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                "@type": "Metadata",
                "dc:contributor": "A.N. Other",
                "dc:title": "The title",
                "dcterms:abstract": "This is my abstract",
            },
            "swordMetadataSourceFormat": "http://purl.org/net/sword/3.0/types/Metadata",
        }


@pytest.mark.parametrize(
    "filename,error_class",
    [
        ("bagit-broken-sha.zip", ValidationFailed),
        ("bagit-no-bagit-txt.zip", ContentMalformed),
        ("bagit-with-fetch.zip", ValidationFailed),
    ],
)
def test_unpack_document_with_broken_bag(
    api, location, filename, error_class, fixtures_path
):
    with api.test_request_context():
        record = SWORDDeposit.create({})
        with open(os.path.join(fixtures_path, filename), "rb") as stream:
            object_version = ObjectVersion.create(
                bucket=record.bucket, key=filename, stream=stream
            )

        packaging = Packaging.for_record_and_name(record, PackagingFormat.SwordBagIt)
        with pytest.raises(error_class):
            packaging.unpack(object_version)


def test_post_service_document_with_incorrect_content_type(
    api, users, location, fixtures_path
):
    with api.test_request_context():
        record = SWORDDeposit.create({})
        with open(os.path.join(fixtures_path, "bagit.zip"), "rb") as stream:
            object_version = ObjectVersion.create(
                bucket=record.bucket,
                key="bagit.zip",
                stream=stream,
                mimetype="application/tar",
            )

        packaging = Packaging.for_record_and_name(record, PackagingFormat.SwordBagIt)
        with pytest.raises(ContentTypeNotAcceptable):
            packaging.unpack(object_version)



import json

import pytest
from sword3common.exceptions import ContentTypeNotAcceptable

from invenio_sword.api import SWORDDeposit
from invenio_sword.metadata import Metadata
from invenio_sword.metadata import SWORDMetadata


class OtherMetadata(Metadata):
    pass


def test_parse_document(metadata_document):
    sword_metadata = SWORDMetadata.from_document(
        metadata_document, content_type="application/ld+json"
    )
    assert "@id" not in sword_metadata.data
    assert sword_metadata.data == {
        "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
        "@type": "Metadata",
        "dc:title": "The title",
        "dcterms:abstract": "This is my abstract",
        "dc:contributor": "A.N. Other",
    }


def test_parse_document_with_wrong_content_type(metadata_document):
    with pytest.raises(ContentTypeNotAcceptable):
        SWORDMetadata.from_document(metadata_document, content_type="text/yaml")


def test_update_record(metadata_document):
    sword_metadata = SWORDMetadata.from_document(
        metadata_document, content_type="application/ld+json"
    )
    record = SWORDDeposit(data={"metadata": {}})
    sword_metadata.update_record_metadata(record)
    assert record["metadata"]["title_statement"]["title"] == "The title"


def test_metadata_bytes_roundtrip(metadata_document):
    sword_metadata = SWORDMetadata.from_document(
        metadata_document, content_type="application/ld+json"
    )
    assert json.loads(bytes(sword_metadata)) == sword_metadata.data


def test_construct_from_dict(metadata_document):
    data = json.load(metadata_document)
    sword_metadata = SWORDMetadata(data)
    assert sword_metadata.data == data


def test_cant_add_to_other_metadata(metadata_document):
    sword_metadata = SWORDMetadata.from_document(
        metadata_document, content_type="application/ld+json"
    )
    other_metadata = OtherMetadata()

    with pytest.raises(TypeError):
        sword_metadata + other_metadata

    with pytest.raises(TypeError):
        other_metadata + sword_metadata

    assert sword_metadata.__add__(other_metadata) == NotImplemented



import io
import json
from http import HTTPStatus

import pytest
from flask import url_for
from flask_security import url_for_security
from invenio_db import db

from invenio_sword.api import SWORDDeposit
from invenio_sword.metadata import SWORDMetadata


def test_get_metadata_document(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        record = SWORDDeposit.create({})
        record.set_metadata({"dc:title": "Deposit title"}, SWORDMetadata)
        record.commit()
        db.session.commit()

        response = client.get("/sword/deposit/{}".format(record.pid.pid_value))
        assert response.status_code == HTTPStatus.OK

        response = client.get("/sword/deposit/{}/metadata".format(record.pid.pid_value))
        assert response.status_code == HTTPStatus.OK
        assert response.is_json
        assert response.json == {
            "@id": "http://localhost/sword/deposit/{}".format(record.pid.pid_value),
            "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
            "dc:title": "Deposit title",
        }


def test_get_metadata_document_when_not_available(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        db.session.commit()

        status_response = client.get("/sword/deposit/{}".format(record.pid.pid_value))
        assert status_response.status_code == HTTPStatus.OK
        status_response = client.get(
            "/sword/deposit/{}/metadata".format(record.pid.pid_value)
        )
        assert status_response.status_code == HTTPStatus.OK


def test_put_metadata_document_without_body(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        db.session.commit()

        response = client.put("/sword/deposit/{}/metadata".format(record.pid.pid_value))
        assert response.status_code == HTTPStatus.BAD_REQUEST


def test_put_metadata_document_invalid_json(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        db.session.commit()

        response = client.put(
            "/sword/deposit/{}/metadata".format(record.pid.pid_value),
            headers={
                "Content-Type": "application/ld+json",
                "Metadata-Format": "http://purl.org/net/sword/3.0/types/Metadata",
            },
        )
        assert response.status_code == HTTPStatus.BAD_REQUEST


def test_put_metadata_document(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        db.session.commit()

        response = client.put(
            "/sword/deposit/{}/metadata".format(record.pid.pid_value),
            headers={
                "Metadata-Format": "http://purl.org/net/sword/3.0/types/Metadata",
                "Content-Type": "application/ld+json",
            },
            data=json.dumps({}),
        )
        assert response.status_code == HTTPStatus.NO_CONTENT

        record = SWORDDeposit.get_record(record.id)
        assert (
            record["swordMetadataSourceFormat"]
            == "http://purl.org/net/sword/3.0/types/Metadata"
        )
        assert any(
            "http://purl.org/net/sword/3.0/terms/formattedMetadata" in link["rel"]
            and link["metadataFormat"] == "http://purl.org/net/sword/3.0/types/Metadata"
            for link in record.get_status_as_jsonld()["links"]
        )


@pytest.mark.parametrize(
    "view_name,status_code,additional_headers",
    [
        ("invenio_sword.depid_metadata", HTTPStatus.NO_CONTENT, {}),
        (
            "invenio_sword.depid_item",
            HTTPStatus.OK,
            {"Content-Disposition": "attachment; metadata=true"},
        ),
    ],
)
def test_post_metadata_document_to_append(
    api, users, location, es, view_name, status_code, additional_headers
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.set_metadata(
            {
                "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                "dc:title": "Some title",
                "dc:subject": "Some subject",
            },
            SWORDMetadata,
        )
        record.commit()
        db.session.commit()

        response = client.post(
            url_for(view_name, pid_value=record.pid.pid_value),
            headers={
                "Metadata-Format": "http://purl.org/net/sword/3.0/types/Metadata",
                "Content-Type": "application/ld+json",
                **additional_headers,
            },
            data=json.dumps(
                {
                    "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                    "dc:subject": "Another subject",
                    "dc:creator": "A person",
                }
            ),
        )
        assert response.status_code == status_code

        record = SWORDDeposit.get_record(record.id)
        assert (
            record["swordMetadataSourceFormat"]
            == "http://purl.org/net/sword/3.0/types/Metadata"
        )
        assert record["swordMetadata"] == {
            "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
            "dc:title": "Some title",
            "dc:subject": "Another subject",
            "dc:creator": "A person",
        }


def test_post_metadata_document_with_additional_metadata_format(
    api, users, location, es, test_metadata_format
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.set_metadata(
            {
                "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
                "dc:title": "Some title",
                "dc:subject": "Some subject",
            },
            SWORDMetadata,
        )
        record.commit()
        db.session.commit()

        response = client.post(
            "/sword/deposit/{}/metadata".format(record.pid.pid_value),
            headers={
                "Metadata-Format": test_metadata_format,
                "Content-Type": "text/plain",
            },
            data=io.BytesIO(b"some metadata"),
        )
        assert response.status_code == HTTPStatus.NO_CONTENT

        record = SWORDDeposit.get_record(record.id)
        # Check nothing changed
        assert (
            record["swordMetadataSourceFormat"]
            == "http://purl.org/net/sword/3.0/types/Metadata"
        )
        assert record["swordMetadata"] == {
            "@context": "https://swordapp.github.io/swordv3/swordv3.jsonld",
            "dc:title": "Some title",
            "dc:subject": "Some subject",
        }
        assert len(
            [
                link
                for link in record.get_status_as_jsonld()["links"]
                if "http://purl.org/net/sword/3.0/terms/formattedMetadata"
                in link["rel"]
            ]
        )


def test_put_metadata_document_with_unsupported_format(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.commit()
        db.session.commit()

        response = client.put(
            "/sword/deposit/{}/metadata".format(record.pid.pid_value),
            headers={
                "Metadata-Format": "http://sword.invalid/Metadata",
                "Content-Type": "application/ld+json",
            },
            data=json.dumps({}),
        )
        assert response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE
        assert response.json["@type"] == "MetadataFormatNotAcceptable"


def test_delete_metadata_document(api, users, location, es):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )
        record = SWORDDeposit.create({})
        record.set_metadata({"dc:title": "Deposit title"}, SWORDMetadata)
        record.commit()
        db.session.commit()

        assert record.get("swordMetadataSourceFormat") is not None
        assert record.get("swordMetadata") is not None

        response = client.delete(
            "/sword/deposit/{}/metadata".format(record.pid.pid_value)
        )
        assert response.status_code == HTTPStatus.NO_CONTENT

        record = SWORDDeposit.get_record(record.id)
        assert record.get("swordMetadataSourceFormat") is None
        assert record.get("swordMetadata") is None



import datetime
import uuid
from urllib.parse import urlparse

import math
from flask import current_app
from marshmallow import Schema
from marshmallow import ValidationError, fields, post_load, validates, validates_schema
from marshmallow import validate
from marshmallow.validate import Range
from werkzeug.exceptions import HTTPException

from sword3common.constants import JSON_LD_CONTEXT, PackagingFormat

__all__ = ["ByReferenceSchema"]


class ByReferenceFileDefinition:
    def __init__(
        self,
        *,
        url: str = None,
        temporary_id: uuid.UUID = None,
        content_disposition: str,
        content_type: str,
        content_length: int = None,
        dereference: bool,
        packaging: str,
        ttl: datetime.datetime = None
    ):
        assert url or temporary_id
        self.url = url
        self.temporary_id = temporary_id
        self.content_disposition = content_disposition
        self.content_type = content_type
        self.content_length = content_length
        self.dereference = dereference
        self.packaging = packaging
        self.ttl = ttl


class _ByReferenceFileSchema(Schema):
    url = fields.Url(data_key="@id", required=True)
    content_disposition = fields.String(data_key="contentDisposition", required=True)
    content_length = fields.Integer(data_key="contentLength", strict=True)
    content_type = fields.String(data_key="contentType", required=True)
    dereference = fields.Boolean(required=True)
    packaging = fields.String(missing=PackagingFormat.Binary)
    ttl = fields.AwareDateTime()

    @validates_schema
    def validate_url(self, data, **kwargs):
        parsed_url = urlparse(data["url"])
        if parsed_url.hostname == self.context["url_adapter"].server_name:
            try:
                rule, rv = self.context["url_adapter"].match(parsed_url.path)
            except HTTPException:
                return None
            if rule == "invenio_sword.temporary_url":
                data["temporary_id"] = rv["temporary_id"]
                del data["url"]

    @post_load
    def make_object(self, data, **kwargs):
        return ByReferenceFileDefinition(**data)


class ByReferenceSchema(Schema):
    jsonld_context = fields.String(
        data_key="@context",
        validate=[validate.OneOf([JSON_LD_CONTEXT])],
        required=True,
    )
    jsonld_type = fields.String(
        data_key="@type", validate=[validate.OneOf(["ByReference"])], required=True,
    )
    files = fields.List(
        fields.Nested(_ByReferenceFileSchema),
        data_key="byReferenceFiles",
        required=True,
    )


class SegmentInitSchema(Schema):
    """For validating parameters on `Content-Disposition: segment-init` headers"""

    # validate parameters are lambdas to defer accessing current_app until we're inside the application context

    size = fields.Integer(
        required=True,
        validate=lambda value: Range(
            current_app.config["FILES_REST_MULTIPART_MAX_PARTS"]
        ),
    )
    segment_count = fields.Integer(
        required=True,
        validate=lambda value: Range(
            1, current_app.config["FILES_REST_MULTIPART_MAX_PARTS"]
        ),
    )
    segment_size = fields.Integer(
        required=True,
        validate=lambda value: Range(
            current_app.config["FILES_REST_MULTIPART_CHUNKSIZE_MIN"],
            current_app.config["FILES_REST_MULTIPART_CHUNKSIZE_MAX"],
        )(value),
    )

    @validates_schema
    def validate_segment_count(self, data, **kwargs):
        if data["segment_count"] != math.ceil(data["size"] / data["segment_size"]):
            raise ValidationError(
                {
                    "segment_count": "Wrong number of segments for given size and segment_size."
                }
            )


class SegmentUploadSchema(Schema):
    segment_number = fields.Integer(required=True)

    def __init__(self, *args, segment_count, **kwargs):
        self._segment_count = segment_count
        super().__init__(*args, **kwargs)

    @validates("segment_number")
    def validate_segment_number(self, value):
        return Range(1, self._segment_count)(value)



import io
import json
import os
from http import HTTPStatus

import pytest
from flask_security import url_for_security
from invenio_files_rest.models import ObjectVersion
from sword3common.exceptions import ContentMalformed
from sword3common.exceptions import ContentTypeNotAcceptable

from invenio_sword.api import SWORDDeposit
from invenio_sword.packaging import SimpleZipPackaging
from invenio_sword.packaging import SWORDBagItPackaging

original_deposit = (
    object()
)  # A sentinel, representing the original deposit in our expected data

packaged_content = {
    "example.svg": {
        "rel": [
            "http://purl.org/net/sword/3.0/terms/derivedResource",
            "http://purl.org/net/sword/3.0/terms/fileSetFile",
        ],
        "contentType": "image/svg+xml",
        "status": "http://purl.org/net/sword/3.0/filestate/ingested",
        "derivedFrom": original_deposit,
    },
    "hello.txt": {
        "rel": [
            "http://purl.org/net/sword/3.0/terms/derivedResource",
            "http://purl.org/net/sword/3.0/terms/fileSetFile",
        ],
        "contentType": "text/plain",
        "status": "http://purl.org/net/sword/3.0/filestate/ingested",
        "derivedFrom": original_deposit,
    },
}

ingest_test_parameters = [
    # A binary deposit
    (
        "binary.svg",
        "http://purl.org/net/sword/3.0/package/Binary",
        "image/svg+xml",
        {
            "binary.svg": {
                "rel": [
                    "http://purl.org/net/sword/3.0/terms/fileSetFile",
                    "http://purl.org/net/sword/3.0/terms/originalDeposit",
                ],
                "contentType": "image/svg+xml",
                "status": "http://purl.org/net/sword/3.0/filestate/ingested",
                "packaging": "http://purl.org/net/sword/3.0/package/Binary",
            }
        },
    ),
    # A simple zip deposit
    (
        "simple.zip",
        "http://purl.org/net/sword/3.0/package/SimpleZip",
        "application/zip",
        {
            original_deposit: {
                "rel": ["http://purl.org/net/sword/3.0/terms/originalDeposit"],
                "contentType": "application/zip",
                "status": "http://purl.org/net/sword/3.0/filestate/ingested",
                "packaging": "http://purl.org/net/sword/3.0/package/SimpleZip",
            },
            **packaged_content,  # type: ignore
        },
    ),
    # A BagIt deposit
    (
        "bagit.zip",
        "http://purl.org/net/sword/3.0/package/SWORDBagIt",
        "application/zip",
        {
            original_deposit: {
                "rel": ["http://purl.org/net/sword/3.0/terms/originalDeposit"],
                "contentType": "application/zip",
                "status": "http://purl.org/net/sword/3.0/filestate/ingested",
                "packaging": "http://purl.org/net/sword/3.0/package/SWORDBagIt",
            },
            **packaged_content,  # type: ignore
        },
    ),
]


@pytest.mark.parametrize(
    "filename,packaging,content_type,expected_links", ingest_test_parameters
)
def test_ingest(
    api,
    users,
    location,
    es,
    task_delay,
    fixtures_path,
    filename,
    packaging,
    content_type,
    expected_links,
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        with open(os.path.join(fixtures_path, filename), "rb") as f:
            response = client.post(
                "/sword/service-document",
                data=f,
                headers={
                    "Packaging": packaging,
                    "Content-Type": content_type,
                    "Content-Disposition": "attachment; filename={}".format(filename),
                },
            )
        assert response.status_code == HTTPStatus.CREATED

        assert task_delay.call_count == 1
        task_self = task_delay.call_args[0][0]
        task_self.apply()

        response = client.get(response.headers["Location"])

        metadata_file_count = 0

        for link in response.json["links"]:
            key = link["@id"].split("/", 7)[-1]
            if key.startswith(".metadata-"):
                metadata_file_count += 1
                continue

            if (
                "http://purl.org/net/sword/3.0/terms/originalDeposit" in link["rel"]
                and "http://purl.org/net/sword/3.0/terms/fileSetFile" not in link["rel"]
            ):
                expected_link = expected_links[original_deposit]
            else:
                expected_link = expected_links[key]

            if "derivedFrom" in link:
                link["derivedFrom"] = original_deposit

            expected_link["@id"] = response.json["@id"] + "/file/" + key

            assert link == expected_link

        assert len(response.json["links"]) == len(expected_links) + metadata_file_count


@pytest.mark.parametrize(
    "filename,packaging,content_type,expected_links", ingest_test_parameters
)
def test_create_with_metadata_and_then_ingest(
    api,
    users,
    location,
    es,
    fixtures_path,
    task_delay,
    filename,
    packaging,
    content_type,
    expected_links,
):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post(
            "/sword/service-document",
            data=io.BytesIO(json.dumps({}).encode()),
            headers={
                "Metadata-Format": "http://purl.org/net/sword/3.0/types/Metadata",
                "Content-Type": "application/ld+json",
                "Content-Disposition": "attachment; metadata=true",
                "In-Progress": "true",
            },
        )
        assert response.status_code == HTTPStatus.CREATED
        assert task_delay.call_count == 0

        with open(os.path.join(fixtures_path, filename), "rb") as f:
            client.post(
                response.headers["Location"],
                data=f,
                headers={
                    "Packaging": packaging,
                    "Content-Type": content_type,
                    "Content-Disposition": "attachment; filename={}".format(filename),
                },
            )

        assert task_delay.call_count == 1
        task_self = task_delay.call_args[0][0]
        task_self.apply()

        response = client.get(response.headers["Location"])

        metadata_link_count = 0

        for link in response.json["links"]:
            key = link["@id"].split("/", 7)[-1]

            # Ignore ingested metadata files
            if key.startswith(".metadata-"):
                metadata_link_count += 1
                continue

            if (
                "http://purl.org/net/sword/3.0/terms/originalDeposit" in link["rel"]
                and "http://purl.org/net/sword/3.0/terms/fileSetFile" not in link["rel"]
            ):
                expected_link = expected_links[original_deposit]
            else:
                expected_link = expected_links[key]

            if "derivedFrom" in link:
                link["derivedFrom"] = original_deposit

            expected_link["@id"] = response.json["@id"] + "/file/" + key

            assert link == expected_link

        assert len(response.json["links"]) == len(expected_links) + metadata_link_count


@pytest.mark.parametrize(
    "filename,content_type,packaging_class,exception_class",
    [
        ("simple.zip", "application/zip", SWORDBagItPackaging, ContentMalformed),
        ("binary.svg", "application/zip", SWORDBagItPackaging, ContentMalformed),
        ("binary.svg", "application/zip", SimpleZipPackaging, ContentMalformed),
        ("binary.svg", "image/svg+xml", SWORDBagItPackaging, ContentTypeNotAcceptable),
        ("binary.svg", "image/svg+xml", SimpleZipPackaging, ContentTypeNotAcceptable),
    ],
)
def test_bad_files(
    api,
    location,
    filename,
    content_type,
    packaging_class,
    exception_class,
    fixtures_path,
):
    with api.app_context():
        record = SWORDDeposit.create({})
        packaging = packaging_class(record)
        with open(os.path.join(fixtures_path, filename), "rb") as stream:
            object_version = ObjectVersion.create(
                record.bucket, key=filename, stream=stream, mimetype=content_type
            )
        with pytest.raises(exception_class):
            packaging.unpack(object_version)


def test_unknown_packaging_format(api, location, users, task_delay):
    with api.test_request_context(), api.test_client() as client:
        client.post(
            url_for_security("login"),
            data={"email": users[0]["email"], "password": "tester"},
        )

        response = client.post(
            "/sword/service-document",
            data=io.BytesIO(b"data"),
            headers={
                "Packaging": "http://something.invalid/",
                "Content-Type": "text/plain",
                "Content-Disposition": "attachment; filename=data.txt",
            },
        )

        assert response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE
        assert response.json["@type"] == "PackagingFormatNotAcceptable"

        assert not task_delay.called



from enum import Enum
from typing import Dict
from typing import Mapping
from typing import Union

from invenio_files_rest.models import ObjectVersion
from invenio_files_rest.models import ObjectVersionTag
from invenio_sword.enum import FileState
from invenio_sword.enum import ObjectTagKey


class TagManager(Dict[ObjectTagKey, Union[str, Enum]]):
    enum_keys = {
        ObjectTagKey.FileState: FileState,
    }

    def __init__(self, object_version: ObjectVersion):
        self._object_version = object_version
        super().__init__(
            {
                ObjectTagKey(key): self.enum_keys.get(ObjectTagKey(key), str)(value)
                for key, value in self._object_version.get_tags().items()
            }
        )

    def __setitem__(self, key: ObjectTagKey, value: Union[str, Enum]):  # type: ignore
        super().__setitem__(key, self.enum_keys.get(ObjectTagKey(key), str)(value))
        if key in self.enum_keys:
            # Check this is a valid value
            self.enum_keys[key](value)
        if isinstance(value, Enum):
            value = value.value
        ObjectVersionTag.create_or_update(self._object_version, key.value, value)

    def __delitem__(self, key: ObjectTagKey):  # type: ignore
        ObjectVersionTag.delete(self._object_version, key.value)
        super().pop(key, None)

    def update(  # type: ignore
        self,
        __m: Mapping[ObjectTagKey, Union[str, Enum]] = None,
        **kwargs: Union[str, Enum]
    ) -> None:
        for mapping_key, value in (__m or {}).items():
            self[mapping_key] = value
        for kwargs_key, value in kwargs.items():
            self[ObjectTagKey(kwargs_key)] = value
